# -*- coding: utf-8 -*-
from django.contrib.admin import FieldListFilter
from django.contrib.admin.utils import prepare_lookup_value
from django.utils.translation import gettext_lazy as _


class NullFieldListFilter(FieldListFilter):
    def __init__(self, field, request, params, model, model_admin, field_path):
        self.lookup_kwarg = "{0}__isnull".format(field_path)
        super().__init__(field, request, params, model, model_admin, field_path)
        lookup_choices = self.lookups(request, model_admin)
        self.lookup_choices = () if lookup_choices is None else list(lookup_choices)

    def expected_parameters(self):
        return [self.lookup_kwarg]

    def value(self):
        return self.used_parameters.get(self.lookup_kwarg, None)

    def lookups(self, request, model_admin):
        return (
            ("1", _("Yes")),
            ("0", _("No")),
        )

    def choices(self, cl):
        yield {
            "selected": self.value() is None,
            "query_string": cl.get_query_string({}, [self.lookup_kwarg]),
            "display": _("All"),
        }
        for lookup, title in self.lookup_choices:
            yield {
                "selected": self.value()
                == prepare_lookup_value(self.lookup_kwarg, lookup),
                "query_string": cl.get_query_string(
                    {
                        self.lookup_kwarg: lookup,
                    },
                    [],
                ),
                "display": title,
            }

    def queryset(self, request, queryset):
        if self.value() is not None:
            kwargs = {self.lookup_kwarg: self.value()}
            return queryset.filter(**kwargs)
        return queryset


class NotNullFieldListFilter(NullFieldListFilter):
    def lookups(self, request, model_admin):
        return (
            ("0", _("Yes")),
            ("1", _("No")),
        )



# -*- coding: utf-8 -*-
from django.core.management import call_command


def test_without_args(capsys):
    call_command("generate_password")

    out, err = capsys.readouterr()
    assert out


def test_with_length_args(capsys):
    length = 20
    call_command("generate_password", length=length)

    out, err = capsys.readouterr()
    assert len(out.rstrip("\n")) == length



# -*- coding: utf-8 -*-
import string
import pytest

from django.test import TestCase

from .testapp.models import (
    RandomCharTestModel,
    RandomCharTestModelUnique,
    RandomCharTestModelLowercase,
)
from .testapp.models import (
    RandomCharTestModelUppercase,
    RandomCharTestModelAlpha,
    RandomCharTestModelDigits,
)
from .testapp.models import (
    RandomCharTestModelPunctuation,
    RandomCharTestModelLowercaseAlphaDigits,
    RandomCharTestModelUppercaseAlphaDigits,
)
from .testapp.models import RandomCharTestModelUniqueTogether

from unittest import mock


class RandomCharFieldTest(TestCase):
    def testRandomCharField(self):
        m = RandomCharTestModel()
        m.save()
        assert len(m.random_char_field) == 8, m.random_char_field

    def testRandomCharFieldUnique(self):
        m = RandomCharTestModelUnique()
        m.save()
        assert len(m.random_char_field) == 8, m.random_char_field

    def testRandomCharFieldLowercase(self):
        m = RandomCharTestModelLowercase()
        m.save()
        for c in m.random_char_field:
            assert c.islower(), m.random_char_field

    def testRandomCharFieldUppercase(self):
        m = RandomCharTestModelUppercase()
        m.save()
        for c in m.random_char_field:
            assert c.isupper(), m.random_char_field

    def testRandomCharFieldAlpha(self):
        m = RandomCharTestModelAlpha()
        m.save()
        for c in m.random_char_field:
            assert c.isalpha(), m.random_char_field

    def testRandomCharFieldDigits(self):
        m = RandomCharTestModelDigits()
        m.save()
        for c in m.random_char_field:
            assert c.isdigit(), m.random_char_field

    def testRandomCharFieldPunctuation(self):
        m = RandomCharTestModelPunctuation()
        m.save()
        for c in m.random_char_field:
            assert c in string.punctuation, m.random_char_field

    def testRandomCharTestModelLowercaseAlphaDigits(self):
        m = RandomCharTestModelLowercaseAlphaDigits()
        m.save()
        for c in m.random_char_field:
            assert c.isdigit() or (c.isalpha() and c.islower()), m.random_char_field

    def testRandomCharTestModelUppercaseAlphaDigits(self):
        m = RandomCharTestModelUppercaseAlphaDigits()
        m.save()
        for c in m.random_char_field:
            assert c.isdigit() or (c.isalpha() and c.isupper()), m.random_char_field

    def testRandomCharTestModelDuplicate(self):
        m = RandomCharTestModelUnique()
        m.save()
        with mock.patch(
            "django_extensions.db.fields.RandomCharField.random_char_generator"
        ) as func:
            func.return_value = iter([m.random_char_field, "aaa"])
            m = RandomCharTestModelUnique()
            m.save()
        assert m.random_char_field == "aaa"

    def testRandomCharTestModelAsserts(self):
        with mock.patch("django_extensions.db.fields.get_random_string") as mock_sample:
            mock_sample.return_value = "aaa"
            m = RandomCharTestModelUnique()
            m.save()

            m = RandomCharTestModelUnique()
            with pytest.raises(RuntimeError):
                m.save()

    def testRandomCharTestModelUniqueTogether(self):
        with mock.patch("django_extensions.db.fields.get_random_string") as mock_sample:
            mock_sample.return_value = "aaa"
            m = RandomCharTestModelUniqueTogether()
            m.common_field = "bbb"
            m.save()

            m = RandomCharTestModelUniqueTogether()
            m.common_field = "bbb"
            with pytest.raises(RuntimeError):
                m.save()



# -*- coding: utf-8 -*-
import pytest
from django.core.management import call_command

from unittest import mock


@pytest.mark.django_db
def test_initialize_runserver_plus():
    with mock.patch(
        "django_extensions.management.commands.runserver_plus.run_simple"
    ) as run_simple:
        call_command("runserver_plus")
        assert run_simple.called, "werkzeug.run_simple was not called"



# -*- coding: utf-8 -*-
"""
Sync Media to S3
================

Django command that scans all files in your settings.MEDIA_ROOT and
settings.STATIC_ROOT folders and uploads them to S3 with the same directory
structure.

This command can optionally do the following but it is off by default:
* gzip compress any CSS and Javascript files it finds and adds the appropriate
  'Content-Encoding' header.
* set a far future 'Expires' header for optimal caching.
* upload only media or static files.
* use any other provider compatible with Amazon S3.
* set other than 'public-read' ACL.

Note: This script requires the Python boto library and valid Amazon Web
Services API keys.

Required settings.py variables:
AWS_ACCESS_KEY_ID = ''
AWS_SECRET_ACCESS_KEY = ''
AWS_BUCKET_NAME = ''

When you call this command with the `--renamegzip` param, it will add
the '.gz' extension to the file name. But Safari just doesn't recognize
'.gz' files and your site won't work on it! To fix this problem, you can
set any other extension (like .jgz) in the `SYNC_S3_RENAME_GZIP_EXT`
variable.

Command options are:
  -p PREFIX, --prefix=PREFIX
                        The prefix to prepend to the path on S3.
  --gzip                Enables gzipping CSS and Javascript files.
  --expires             Enables setting a far future expires header.
  --force               Skip the file mtime check to force upload of all
                        files.
  --filter-list         Override default directory and file exclusion
                        filters. (enter as comma separated line)
  --renamegzip          Enables renaming of gzipped files by appending '.gz'.
                        to the original file name. This way your original
                        assets will not be replaced by the gzipped ones.
                        You can change the extension setting the
                        `SYNC_S3_RENAME_GZIP_EXT` var in your settings.py
                        file.
  --invalidate          Invalidates the objects in CloudFront after uploading
                        stuff to s3.
  --media-only          Only MEDIA_ROOT files will be uploaded to S3.
  --static-only         Only STATIC_ROOT files will be uploaded to S3.
  --s3host              Override default s3 host.
  --acl                 Override default ACL settings ('public-read' if
                        settings.AWS_DEFAULT_ACL is not defined).

TODO:
 * Use fnmatch (or regex) to allow more complex FILTER_LIST rules.

"""

import datetime
import email
import gzip
import mimetypes
import os
import time
from typing import List  # NOQA

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from io import StringIO

from django_extensions.management.utils import signalcommand


try:
    import boto
except ImportError:
    HAS_BOTO = False
else:
    HAS_BOTO = True


class Command(BaseCommand):
    # Extra variables to avoid passing these around
    AWS_ACCESS_KEY_ID = ""
    AWS_SECRET_ACCESS_KEY = ""
    AWS_BUCKET_NAME = ""
    AWS_CLOUDFRONT_DISTRIBUTION = ""
    SYNC_S3_RENAME_GZIP_EXT = ""

    DIRECTORIES = ""
    FILTER_LIST = [".DS_Store", ".svn", ".hg", ".git", "Thumbs.db"]
    GZIP_CONTENT_TYPES = (
        "text/css",
        "application/javascript",
        "application/x-javascript",
        "text/javascript",
    )

    uploaded_files = []  # type: List[str]
    upload_count = 0
    skip_count = 0

    help = "Syncs the complete MEDIA_ROOT structure and files to S3 into the given bucket name."  # noqa: E501
    args = "bucket_name"

    can_import_settings = True

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "-p",
            "--prefix",
            dest="prefix",
            default=getattr(settings, "SYNC_S3_PREFIX", ""),
            help="The prefix to prepend to the path on S3.",
        )
        parser.add_argument(
            "-d", "--dir", dest="dir", help="Custom static root directory to use"
        )
        parser.add_argument(
            "--s3host",
            dest="s3host",
            default=getattr(settings, "AWS_S3_HOST", ""),
            help="The s3 host (enables connecting to other providers/regions)",
        )
        parser.add_argument(
            "--acl",
            dest="acl",
            default=getattr(settings, "AWS_DEFAULT_ACL", "public-read"),
            help="Enables to override default acl (public-read).",
        )
        parser.add_argument(
            "--gzip",
            action="store_true",
            dest="gzip",
            default=False,
            help="Enables gzipping CSS and Javascript files.",
        )
        parser.add_argument(
            "--renamegzip",
            action="store_true",
            dest="renamegzip",
            default=False,
            help=(
                "Enables renaming of gzipped assets to have '.gz' "
                "appended to the filename."
            ),
        )
        parser.add_argument(
            "--expires",
            action="store_true",
            dest="expires",
            default=False,
            help="Enables setting a far future expires header.",
        )
        parser.add_argument(
            "--force",
            action="store_true",
            dest="force",
            default=False,
            help="Skip the file mtime check to force upload of all files.",
        )
        parser.add_argument(
            "--filter-list",
            dest="filter_list",
            action="store",
            default="",
            help=(
                "Override default directory and file exclusion filters. "
                "(enter as comma separated line)"
            ),
        )
        parser.add_argument(
            "--invalidate",
            dest="invalidate",
            default=False,
            action="store_true",
            help="Invalidates the associated objects in CloudFront",
        )
        parser.add_argument(
            "--media-only",
            dest="media_only",
            default="",
            action="store_true",
            help="Only MEDIA_ROOT files will be uploaded to S3",
        )
        parser.add_argument(
            "--static-only",
            dest="static_only",
            default="",
            action="store_true",
            help="Only STATIC_ROOT files will be uploaded to S3",
        )

    @signalcommand
    def handle(self, *args, **options):
        if not HAS_BOTO:
            raise CommandError(
                "Please install the 'boto' Python library. ($ pip install boto)"
            )

        # Check for AWS keys in settings
        if not hasattr(settings, "AWS_ACCESS_KEY_ID") or not hasattr(
            settings, "AWS_SECRET_ACCESS_KEY"
        ):
            raise CommandError(
                (
                    "Missing AWS keys from settings file. Please supply both "
                    "AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY."
                )
            )
        else:
            self.AWS_ACCESS_KEY_ID = settings.AWS_ACCESS_KEY_ID
            self.AWS_SECRET_ACCESS_KEY = settings.AWS_SECRET_ACCESS_KEY

        if not hasattr(settings, "AWS_BUCKET_NAME"):
            raise CommandError(
                (
                    "Missing bucket name from settings file. Please add the "
                    "AWS_BUCKET_NAME to your settings file."
                )
            )
        else:
            if not settings.AWS_BUCKET_NAME:
                raise CommandError("AWS_BUCKET_NAME cannot be empty.")
        self.AWS_BUCKET_NAME = settings.AWS_BUCKET_NAME

        if not hasattr(settings, "MEDIA_ROOT"):
            raise CommandError("MEDIA_ROOT must be set in your settings.")
        else:
            if not settings.MEDIA_ROOT:
                raise CommandError("MEDIA_ROOT must be set in your settings.")

        self.AWS_CLOUDFRONT_DISTRIBUTION = getattr(
            settings, "AWS_CLOUDFRONT_DISTRIBUTION", ""
        )

        self.SYNC_S3_RENAME_GZIP_EXT = getattr(
            settings, "SYNC_S3_RENAME_GZIP_EXT", ".gz"
        )

        self.verbosity = options["verbosity"]
        self.prefix = options["prefix"]
        self.do_gzip = options["gzip"]
        self.rename_gzip = options["renamegzip"]
        self.do_expires = options["expires"]
        self.do_force = options["force"]
        self.invalidate = options["invalidate"]
        self.DIRECTORIES = options["dir"]
        self.s3host = options["s3host"]
        self.default_acl = options["acl"]
        self.FILTER_LIST = getattr(settings, "FILTER_LIST", self.FILTER_LIST)
        filter_list = options["filter_list"]
        if filter_list:
            # command line option overrides default filter_list and
            # settings.filter_list
            self.FILTER_LIST = filter_list.split(",")

        self.media_only = options["media_only"]
        self.static_only = options["static_only"]
        # Get directories
        if self.media_only and self.static_only:
            raise CommandError(
                "Can't use --media-only and --static-only together. "
                "Better not use anything..."
            )
        elif self.media_only:
            self.DIRECTORIES = [settings.MEDIA_ROOT]
        elif self.static_only:
            self.DIRECTORIES = [settings.STATIC_ROOT]
        elif self.DIRECTORIES:
            self.DIRECTORIES = [self.DIRECTORIES]
        else:
            self.DIRECTORIES = [settings.MEDIA_ROOT, settings.STATIC_ROOT]

        # Now call the syncing method to walk the MEDIA_ROOT directory and
        # upload all files found.
        self.sync_s3()

        # Sending the invalidation request to CloudFront if the user
        # requested this action
        if self.invalidate:
            self.invalidate_objects_cf()

        print("")
        print("%d files uploaded." % self.upload_count)
        print("%d files skipped." % self.skip_count)

    def open_cf(self):
        """Return an open connection to CloudFront"""
        return boto.connect_cloudfront(
            self.AWS_ACCESS_KEY_ID, self.AWS_SECRET_ACCESS_KEY
        )

    def invalidate_objects_cf(self):
        """Split the invalidation request in groups of 1000 objects"""
        if not self.AWS_CLOUDFRONT_DISTRIBUTION:
            raise CommandError(
                "An object invalidation was requested but the variable "
                "AWS_CLOUDFRONT_DISTRIBUTION is not present in your settings."
            )

        # We can't send more than 1000 objects in the same invalidation
        # request.
        chunk = 1000

        # Connecting to CloudFront
        conn = self.open_cf()

        # Splitting the object list
        objs = self.uploaded_files
        chunks = [objs[i : i + chunk] for i in range(0, len(objs), chunk)]

        # Invalidation requests
        for paths in chunks:
            conn.create_invalidation_request(self.AWS_CLOUDFRONT_DISTRIBUTION, paths)

    def sync_s3(self):
        """Walk the media/static directories and syncs files to S3"""
        bucket, key = self.open_s3()
        for directory in self.DIRECTORIES:
            for root, dirs, files in os.walk(directory):
                self.upload_s3(
                    (bucket, key, self.AWS_BUCKET_NAME, directory), root, files, dirs
                )

    def compress_string(self, s):
        """Gzip a given string."""
        zbuf = StringIO()
        zfile = gzip.GzipFile(mode="wb", compresslevel=6, fileobj=zbuf)
        zfile.write(s)
        zfile.close()
        return zbuf.getvalue()

    def get_s3connection_kwargs(self):
        """Return connection kwargs as a dict"""
        kwargs = {}
        if self.s3host:
            kwargs["host"] = self.s3host
        return kwargs

    def open_s3(self):
        """Open connection to S3 returning bucket and key"""
        conn = boto.connect_s3(
            self.AWS_ACCESS_KEY_ID,
            self.AWS_SECRET_ACCESS_KEY,
            **self.get_s3connection_kwargs(),
        )
        try:
            bucket = conn.get_bucket(self.AWS_BUCKET_NAME)
        except boto.exception.S3ResponseError:
            bucket = conn.create_bucket(self.AWS_BUCKET_NAME)
        return bucket, boto.s3.key.Key(bucket)

    def upload_s3(self, arg, dirname, names, dirs):
        bucket, key, bucket_name, root_dir = arg

        # Skip directories we don't want to sync
        if (
            os.path.basename(dirname) in self.FILTER_LIST
            and os.path.dirname(dirname) in self.DIRECTORIES
        ):
            # prevent walk from processing subfiles/subdirs below the ignored one
            del dirs[:]
            return

        # Later we assume the MEDIA_ROOT ends with a trailing slash
        if not root_dir.endswith(os.path.sep):
            root_dir = root_dir + os.path.sep

        for file in names:
            headers = {}

            if file in self.FILTER_LIST:
                continue  # Skip files we don't want to sync

            filename = os.path.join(dirname, file)
            if os.path.isdir(filename):
                continue  # Don't try to upload directories

            file_key = filename[len(root_dir) :]
            if self.prefix:
                file_key = "%s/%s" % (self.prefix, file_key)

            # Check if file on S3 is older than local file, if so, upload
            if not self.do_force:
                s3_key = bucket.get_key(file_key)
                if s3_key:
                    s3_datetime = datetime.datetime(
                        *time.strptime(
                            s3_key.last_modified, "%a, %d %b %Y %H:%M:%S %Z"
                        )[0:6]
                    )
                    local_datetime = datetime.datetime.utcfromtimestamp(
                        os.stat(filename).st_mtime
                    )
                    if local_datetime < s3_datetime:
                        self.skip_count += 1
                        if self.verbosity > 1:
                            print(
                                "File %s hasn't been modified since last being uploaded"
                                % file_key
                            )
                        continue

            # File is newer, let's process and upload
            if self.verbosity > 0:
                print("Uploading %s..." % file_key)

            content_type = mimetypes.guess_type(filename)[0]
            if content_type:
                headers["Content-Type"] = content_type
            else:
                headers["Content-Type"] = "application/octet-stream"

            file_obj = open(filename, "rb")
            file_size = os.fstat(file_obj.fileno()).st_size
            filedata = file_obj.read()
            if self.do_gzip:
                # Gzip only if file is large enough (>1K is recommended)
                # and only if file is a common text type (not a binary file)
                if file_size > 1024 and content_type in self.GZIP_CONTENT_TYPES:
                    filedata = self.compress_string(filedata)
                    if self.rename_gzip:
                        # If rename_gzip is True, then rename the file
                        # by appending an extension (like '.gz)' to
                        # original filename.
                        file_key = "%s.%s" % (file_key, self.SYNC_S3_RENAME_GZIP_EXT)
                    headers["Content-Encoding"] = "gzip"
                    if self.verbosity > 1:
                        print(
                            "\tgzipped: %dk to %dk"
                            % (file_size / 1024, len(filedata) / 1024)
                        )
            if self.do_expires:
                # HTTP/1.0
                headers["Expires"] = "%s GMT" % (
                    email.Utils.formatdate(
                        time.mktime(
                            (
                                datetime.datetime.now()
                                + datetime.timedelta(days=365 * 2)
                            ).timetuple()
                        )
                    )
                )
                # HTTP/1.1
                headers["Cache-Control"] = "max-age %d" % (3600 * 24 * 365 * 2)
                if self.verbosity > 1:
                    print("\texpires: %s" % headers["Expires"])
                    print("\tcache-control: %s" % headers["Cache-Control"])

            try:
                key.name = file_key
                key.set_contents_from_string(
                    filedata, headers, replace=True, policy=self.default_acl
                )
            except boto.exception.S3CreateError as e:
                print("Failed: %s" % e)
            except Exception as e:
                print(e)
                raise
            else:
                self.upload_count += 1
                self.uploaded_files.append(file_key)

            file_obj.close()



# -*- coding: utf-8 -*-
"""
runprofileserver.py

    Starts a lightweight Web server with profiling enabled.

Credits for kcachegrind support taken from lsprofcalltree.py go to:
 David Allouche
 Jp Calderone & Itamar Shtull-Trauring
 Johan Dahlin
"""

import sys
from datetime import datetime

from django.conf import settings
from django.contrib.staticfiles.handlers import StaticFilesHandler
from django.core.management.base import BaseCommand, CommandError
from django.core.servers.basehttp import get_internal_wsgi_application

from django_extensions.management.utils import signalcommand

USE_STATICFILES = "django.contrib.staticfiles" in settings.INSTALLED_APPS


class KCacheGrind:
    def __init__(self, profiler):
        self.data = profiler.getstats()
        self.out_file = None

    def output(self, out_file):
        self.out_file = out_file
        self.out_file.write("events: Ticks\n")
        self._print_summary()
        for entry in self.data:
            self._entry(entry)

    def _print_summary(self):
        max_cost = 0
        for entry in self.data:
            totaltime = int(entry.totaltime * 1000)
            max_cost = max(max_cost, totaltime)
        self.out_file.write("summary: %d\n" % (max_cost,))

    def _entry(self, entry):
        out_file = self.out_file

        code = entry.code
        if isinstance(code, str):
            out_file.write("fn=%s\n" % code)
        else:
            out_file.write("fl=%s\n" % code.co_filename)
            out_file.write("fn=%s\n" % code.co_name)

        inlinetime = int(entry.inlinetime * 1000)
        if isinstance(code, str):
            out_file.write("0  %s\n" % inlinetime)
        else:
            out_file.write("%d %d\n" % (code.co_firstlineno, inlinetime))

        # recursive calls are counted in entry.calls
        if entry.calls:
            calls = entry.calls
        else:
            calls = []

        if isinstance(code, str):
            lineno = 0
        else:
            lineno = code.co_firstlineno

        for subentry in calls:
            self._subentry(lineno, subentry)
        out_file.write("\n")

    def _subentry(self, lineno, subentry):
        out_file = self.out_file
        code = subentry.code
        if isinstance(code, str):
            out_file.write("cfn=%s\n" % code)
            out_file.write("calls=%d 0\n" % (subentry.callcount,))
        else:
            out_file.write("cfl=%s\n" % code.co_filename)
            out_file.write("cfn=%s\n" % code.co_name)
            out_file.write("calls=%d %d\n" % (subentry.callcount, code.co_firstlineno))

        totaltime = int(subentry.totaltime * 1000)
        out_file.write("%d %d\n" % (lineno, totaltime))


class Command(BaseCommand):
    help = "Starts a lightweight Web server with profiling enabled."
    args = "[optional port number, or ipaddr:port]"

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "addrport", nargs="?", help="Optional port number, or ipaddr:port"
        )
        parser.add_argument(
            "--noreload",
            action="store_false",
            dest="use_reloader",
            default=True,
            help="Tells Django to NOT use the auto-reloader.",
        )
        parser.add_argument(
            "--nothreading",
            action="store_false",
            dest="use_threading",
            default=True,
            help="Tells Django to NOT use threading.",
        )
        parser.add_argument(
            "--prof-path",
            dest="prof_path",
            default="/tmp",
            help="Specifies the directory which to save profile information in.",
        )
        parser.add_argument(
            "--prof-file",
            dest="prof_file",
            default="{path}.{duration:06d}ms.{time}",
            help='Set filename format, default if "{path}.{duration:06d}ms.{time}".',
        )
        parser.add_argument(
            "--nomedia",
            action="store_true",
            dest="no_media",
            default=False,
            help="Do not profile MEDIA_URL",
        )
        parser.add_argument(
            "--kcachegrind",
            action="store_true",
            dest="use_lsprof",
            default=False,
            help="Create kcachegrind compatible lsprof files, this requires "
            "and automatically enables cProfile.",
        )

        if USE_STATICFILES:
            parser.add_argument(
                "--nostatic",
                action="store_false",
                dest="use_static_handler",
                default=True,
                help="Tells Django to NOT automatically serve static files "
                "at STATIC_URL.",
            )
            parser.add_argument(
                "--insecure",
                action="store_true",
                dest="insecure_serving",
                default=False,
                help="Allows serving static files even if DEBUG is False.",
            )

    @signalcommand
    def handle(self, addrport="", *args, **options):
        import django
        import socket
        import errno
        from django.core.servers.basehttp import run

        if not addrport:
            addr = ""
            port = "8000"
        else:
            try:
                addr, port = addrport.split(":")
            except ValueError:
                addr, port = "", addrport
        if not addr:
            addr = "127.0.0.1"

        if not port.isdigit():
            raise CommandError("%r is not a valid port number." % port)

        use_reloader = options["use_reloader"]
        shutdown_message = options.get("shutdown_message", "")
        no_media = options["no_media"]
        quit_command = (sys.platform == "win32") and "CTRL-BREAK" or "CONTROL-C"

        def inner_run():
            import os
            import time

            import cProfile

            USE_LSPROF = options["use_lsprof"]

            prof_path = options["prof_path"]

            prof_file = options["prof_file"]
            if not prof_file.format(path="1", duration=2, time=3):
                prof_file = "{path}.{duration:06d}ms.{time}"
                print(
                    "Filename format is wrong. "
                    "Default format used: '{path}.{duration:06d}ms.{time}'."
                )

            def get_exclude_paths():
                exclude_paths = []
                media_url = getattr(settings, "MEDIA_URL", None)
                if media_url:
                    exclude_paths.append(media_url)
                static_url = getattr(settings, "STATIC_URL", None)
                if static_url:
                    exclude_paths.append(static_url)
                return exclude_paths

            def make_profiler_handler(inner_handler):
                def handler(environ, start_response):
                    path_info = environ["PATH_INFO"]
                    # when using something like a dynamic site middleware is could be
                    # necessary to refetch the exclude_paths every time since they could
                    # change per site.
                    if no_media and any(
                        path_info.startswith(p) for p in get_exclude_paths()
                    ):
                        return inner_handler(environ, start_response)
                    path_name = path_info.strip("/").replace("/", ".") or "root"
                    profname = "%s.%d.prof" % (path_name, time.time())
                    profname = os.path.join(prof_path, profname)
                    prof = cProfile.Profile()
                    start = datetime.now()
                    try:
                        return prof.runcall(inner_handler, environ, start_response)
                    finally:
                        # seeing how long the request took is important!
                        elap = datetime.now() - start
                        elapms = elap.seconds * 1000.0 + elap.microseconds / 1000.0
                        if USE_LSPROF:
                            kg = KCacheGrind(prof)
                            with open(profname, "w") as f:
                                kg.output(f)
                        else:
                            prof.dump_stats(profname)
                        profname2 = prof_file.format(
                            path=path_name, duration=int(elapms), time=int(time.time())
                        )
                        profname2 = os.path.join(prof_path, "%s.prof" % profname2)
                        os.rename(profname, profname2)

                return handler

            print("Performing system checks...")
            self.check(display_num_errors=True)

            print(
                "\nDjango version %s, using settings %r"
                % (django.get_version(), settings.SETTINGS_MODULE)
            )
            print("Development server is running at http://%s:%s/" % (addr, port))
            print("Quit the server with %s." % quit_command)
            try:
                handler = get_internal_wsgi_application()
                if USE_STATICFILES:
                    use_static_handler = options["use_static_handler"]
                    insecure_serving = options["insecure_serving"]
                    if use_static_handler and (settings.DEBUG or insecure_serving):
                        handler = StaticFilesHandler(handler)
                handler = make_profiler_handler(handler)
                run(addr, int(port), handler, threading=options["use_threading"])
            except socket.error as e:
                # Use helpful error messages instead of ugly tracebacks.
                ERRORS = {
                    errno.EACCES: "You don't have permission to access that port.",
                    errno.EADDRINUSE: "That port is already in use.",
                    errno.EADDRNOTAVAIL: "That IP address can't be assigned-to.",
                }
                try:
                    error_text = ERRORS[e.errno]
                except (AttributeError, KeyError):
                    error_text = str(e)
                sys.stderr.write(self.style.ERROR("Error: %s" % error_text) + "\n")
                # Need to use an OS exit because sys.exit doesn't work in a thread
                os._exit(1)
            except KeyboardInterrupt:
                if shutdown_message:
                    print(shutdown_message)
                sys.exit(0)

        if use_reloader:
            try:
                from django.utils.autoreload import run_with_reloader

                run_with_reloader(inner_run)
            except ImportError:
                from django.utils import autoreload

                autoreload.main(inner_run)
        else:
            inner_run()



# -*- coding: utf-8 -*-
from io import BytesIO

import csv
import codecs
import importlib

from django.conf import settings


#
# Django compatibility
#
def load_tag_library(libname):
    """
    Load a templatetag library on multiple Django versions.

    Returns None if the library isn't loaded.
    """
    from django.template.backends.django import get_installed_libraries
    from django.template.library import InvalidTemplateLibrary

    try:
        lib = get_installed_libraries()[libname]
        lib = importlib.import_module(lib).register
        return lib
    except (InvalidTemplateLibrary, KeyError):
        return None


def get_template_setting(template_key, default=None):
    """Read template settings"""
    templates_var = getattr(settings, "TEMPLATES", None)
    if templates_var:
        for tdict in templates_var:
            if template_key in tdict:
                return tdict[template_key]
    return default


class UnicodeWriter:
    """
    CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    We are using this custom UnicodeWriter for python versions 2.x
    """

    def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
        self.queue = BytesIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwds)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writerow(self, row):
        self.writer.writerow([s.encode("utf-8") for s in row])
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        data = data.decode("utf-8")
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)



# -*- coding: utf-8 -*-
import os
import re
import pytest
import inspect

from textwrap import dedent
from django.core.management import call_command
from django.db.models import Model
from django.test import override_settings

from django_extensions.management.commands import shell_plus


def test_shell_plus_command(capsys):
    script = dedent(
        """\
        def _fn(x): return x * 2
        print([_fn(i) for i in range(10)])
        """
    )
    call_command("shell_plus", "--command=" + script)
    out, err = capsys.readouterr()

    assert out.rstrip().endswith(repr([i * 2 for i in range(10)]))


@pytest.mark.django_db()
@override_settings(SHELL_PLUS_SQLPARSE_ENABLED=False, SHELL_PLUS_PYGMENTS_ENABLED=False)
def test_shell_plus_print_sql(capsys):
    try:
        from django.db import connection
        from django.db.backends import utils

        CursorDebugWrapper = utils.CursorDebugWrapper
        force_debug_cursor = True if connection.force_debug_cursor else False
        call_command(
            "shell_plus",
            "--plain",
            "--print-sql",
            "--command=User.objects.all().exists()",
        )
    finally:
        utils.CursorDebugWrapper = CursorDebugWrapper
        connection.force_debug_cursor = force_debug_cursor

    out, err = capsys.readouterr()

    assert re.search(r"SELECT\s+.+\s+FROM\s+.auth_user.\s+LIMIT\s+1", out)


@pytest.mark.django_db()
@override_settings(SHELL_PLUS_SQLPARSE_ENABLED=False, SHELL_PLUS_PYGMENTS_ENABLED=False)
def test_shell_plus_print_sql_truncate(capsys):
    try:
        from django.db import connection
        from django.db.backends import utils

        CursorDebugWrapper = utils.CursorDebugWrapper
        force_debug_cursor = True if connection.force_debug_cursor else False
        call_command(
            "shell_plus",
            "--plain",
            "--print-sql",
            "--truncate-sql=0",
            "--command=User.objects.all().exists()",
        )
    finally:
        utils.CursorDebugWrapper = CursorDebugWrapper
        connection.force_debug_cursor = force_debug_cursor

    out, err = capsys.readouterr()

    assert re.search(r"SELECT\s+.+\s+FROM\s+.auth_user.\s+LIMIT\s+1", out)

    try:
        from django.db import connection
        from django.db.backends import utils

        CursorDebugWrapper = utils.CursorDebugWrapper
        force_debug_cursor = True if connection.force_debug_cursor else False
        call_command(
            "shell_plus",
            "--plain",
            "--print-sql",
            "--truncate-sql=4",
            "--command=User.objects.all().exists()",
        )
    finally:
        utils.CursorDebugWrapper = CursorDebugWrapper
        connection.force_debug_cursor = force_debug_cursor

    out, err = capsys.readouterr()

    assert re.search(r"SELE", out)
    assert not re.search(r"SELEC", out)


def test_shell_plus_plain_startup():
    command = shell_plus.Command()
    command.tests_mode = True

    parser = command.create_parser("test", "shell_plus")
    args = ["--plain"]
    options = parser.parse_args(args=args)

    retcode = command.handle(**vars(options))

    assert retcode == 130


def test_shell_plus_plain_startup_with_pythonrc(monkeypatch):
    command = shell_plus.Command()
    command.tests_mode = True

    parser = command.create_parser("test", "shell_plus")
    args = ["--plain", "--use-pythonrc"]
    options = parser.parse_args(args=args)

    tests_dir = os.path.dirname(
        os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
    )
    pythonrc_file = os.path.join(tests_dir, "pythonrc.py")
    assert os.path.isfile(pythonrc_file)
    monkeypatch.setenv("PYTHONSTARTUP", pythonrc_file)

    retcode = command.handle(**vars(options))
    assert retcode == 130

    imported_objects = command.tests_imported_objects
    assert "pythonrc_test_func" in imported_objects
    assert imported_objects["pythonrc_test_func"]() == "pythonrc was loaded"


def test_shell_plus_plain_loading_standard_django_imports(monkeypatch):
    command = shell_plus.Command()
    command.tests_mode = True

    parser = command.create_parser("test", "shell_plus")
    args = ["--plain"]
    options = parser.parse_args(args=args)

    retcode = command.handle(**vars(options))
    assert retcode == 130

    imported_objects = command.tests_imported_objects
    assert "get_user_model" in imported_objects
    assert "settings" in imported_objects
    assert "timezone" in imported_objects


def test_shell_plus_plain_loading_django_extensions_modules(monkeypatch):
    command = shell_plus.Command()
    command.tests_mode = True

    parser = command.create_parser("test", "shell_plus")
    args = ["--plain"]
    options = parser.parse_args(args=args)

    retcode = command.handle(**vars(options))
    assert retcode == 130

    imported_objects = command.tests_imported_objects
    assert "Club" in imported_objects
    assert "UniqueTestAppModel" in imported_objects
    assert "RandomCharTestModel" in imported_objects


def assert_should_models_be_imported(should_be, cli_arguments=None):
    command = shell_plus.Command()
    objs = command.get_imported_objects(cli_arguments or {})
    imported_models = filter(
        lambda imported: inspect.isclass(imported) and issubclass(imported, Model),
        objs.values(),
    )
    assert bool(list(imported_models)) == should_be


def test_shell_plus_loading_models():
    assert_should_models_be_imported(True)


def test_shell_plus_skipping_models_import_cli():
    assert_should_models_be_imported(False, cli_arguments={"dont_load": ["*"]})


@override_settings(SHELL_PLUS_DONT_LOAD=["*"])
def test_shell_plus_skipping_models_import_settings():
    assert_should_models_be_imported(False)



# -*- coding: utf-8 -*-
from django.test import TestCase
from django_extensions.db.fields.json import dumps, loads, JSONField, JSONDict, JSONList
from .testapp.models import JSONFieldTestModel


class JsonFieldTest(TestCase):
    def test_char_field_create(self):
        j = JSONFieldTestModel.objects.create(a=6, j_field=dict(foo="bar"))
        self.assertEqual(j.a, 6)
        self.assertEqual(j.j_field, {"foo": "bar"})

    def test_char_field_get_or_create(self):
        j, created = JSONFieldTestModel.objects.get_or_create(
            a=6, j_field=dict(foo="bar")
        )

        self.assertTrue(created)
        self.assertEqual(j.a, 6)
        self.assertEqual(j.j_field, {"foo": "bar"})

        j, created = JSONFieldTestModel.objects.get_or_create(
            a=6, j_field=dict(foo="bar")
        )

        self.assertFalse(created)
        self.assertEqual(j.a, 6)
        self.assertEqual(j.j_field, {"foo": "bar"})

    def test_default(self):
        j = JSONFieldTestModel.objects.create(a=1)
        self.assertEqual(j.j_field, {})

    def test_default_mutable(self):
        j1 = JSONFieldTestModel.objects.create(a=1)
        self.assertEqual(j1.j_field, {})

        j2 = JSONFieldTestModel.objects.create(a=1)
        self.assertEqual(j2.j_field, {})

        self.assertIsNot(j1.j_field, j2.j_field)

    def test_get_default(self):
        j_field = JSONField()
        value = j_field.get_default()
        self.assertEqual(value, {})
        self.assertIsInstance(value, JSONDict)

        j_field = JSONField(default={})
        value = j_field.get_default()
        self.assertEqual(value, {})
        self.assertIsInstance(value, JSONDict)

        j_field = JSONField(default="{}")
        value = j_field.get_default()
        self.assertEqual(value, {})
        self.assertIsInstance(value, JSONDict)

        j_field = JSONField(default=[{}])
        value = j_field.get_default()
        self.assertEqual(value, [{}])
        self.assertIsInstance(value, JSONList)

        j_field = JSONField(default="[{}]")
        value = j_field.get_default()
        self.assertEqual(value, [{}])
        self.assertIsInstance(value, JSONList)

        j_field = JSONField(default=lambda: "{}")
        value = j_field.get_default()
        self.assertEqual(value, {})
        self.assertIsInstance(value, JSONDict)

    def test_empty_list(self):
        j = JSONFieldTestModel.objects.create(a=6, j_field=[])
        self.assertIsInstance(j.j_field, list)
        self.assertEqual(j.j_field, [])

    def test_float_values(self):
        """Tests that float values in JSONFields are correctly serialized over
        repeated saves. Regression test for c382398b, which fixes floats
        being returned as strings after a second save."""

        test_instance = JSONFieldTestModel(a=6, j_field={"test": 0.1})
        test_instance.save()

        test_instance = JSONFieldTestModel.objects.get()
        test_instance.save()

        test_instance = JSONFieldTestModel.objects.get()
        self.assertEqual(test_instance.j_field["test"], 0.1)

    def test_get_prep_value(self):
        j_field = JSONField()

        self.assertEqual(
            str(dumps([{"a": "a"}])),
            j_field.get_prep_value(value=[{"a": "a"}]),
        )

        self.assertEqual(
            str(dumps([{"a": "a"}])),
            j_field.get_prep_value(value='[{"a": "a"}]'),
        )

    def test_get_db_prep_save(self):
        j_field = JSONField()

        self.assertEqual(
            str(dumps([{"a": "a"}])),
            j_field.get_db_prep_save(value=[{"a": "a"}], connection=None),
        )

        self.assertEqual(
            str('[{"a": "a"}]'),
            j_field.get_db_prep_save(value='[{"a": "a"}]', connection=None),
        )

    def test_to_python(self):
        j_field = JSONField()

        self.assertEqual(loads("1"), j_field.to_python("1"))

        self.assertEqual(loads('"1"'), j_field.to_python('"1"'))

        self.assertEqual(loads('[{"a": 1}]'), j_field.to_python('[{"a": 1}]'))

        self.assertEqual(loads('[{"a": "1"}]'), j_field.to_python('[{"a": "1"}]'))

    def test_bulk_update_custom_get_prep_value(self):
        obj = JSONFieldTestModel.objects.create(a=1, j_field={"version": "1"})
        obj.j_field["version"] = "1-alpha"
        JSONFieldTestModel.objects.bulk_update([obj], ["j_field"])
        self.assertSequenceEqual(
            JSONFieldTestModel.objects.values("j_field"),
            [{"j_field": {"version": "1-alpha"}}],
        )



# -*- coding: utf-8 -*-
from django_extensions.management.base import LoggingBaseCommand


class Command(LoggingBaseCommand):
    help = "Test error"

    def handle(self, *args, **options):
        raise Exception("Test Error")



# -*- coding: utf-8 -*-
"""
set_fake_passwords.py

    Reset all user passwords to a common value. Useful for testing in a
    development environment. As such, this command is only available when
    setting.DEBUG is True.

"""

from typing import List

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand, CommandError

from django_extensions.management.utils import signalcommand

DEFAULT_FAKE_PASSWORD = "password"


class Command(BaseCommand):
    help = 'DEBUG only: sets all user passwords to a common value ("%s" by default)' % (
        DEFAULT_FAKE_PASSWORD,
    )
    requires_system_checks: List[str] = []

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--prompt",
            dest="prompt_passwd",
            default=False,
            action="store_true",
            help="Prompts for the new password to apply to all users",
        )
        parser.add_argument(
            "--password",
            dest="default_passwd",
            default=DEFAULT_FAKE_PASSWORD,
            help="Use this as default password.",
        )

    @signalcommand
    def handle(self, *args, **options):
        if not settings.DEBUG:
            raise CommandError("Only available in debug mode")

        if options["prompt_passwd"]:
            from getpass import getpass

            passwd = getpass("Password: ")
            if not passwd:
                raise CommandError("You must enter a valid password")
        else:
            passwd = options["default_passwd"]

        User = get_user_model()
        user = User()
        user.set_password(passwd)
        count = User.objects.all().update(password=user.password)

        print("Reset %d passwords" % count)



# -*- coding: utf-8 -*-
import os

import pytest
from unittest.mock import patch

from django.core.management import CommandError, call_command
from django.db import models
from django.test import TestCase, override_settings
from tests import testapp_with_appconfig

MIGRATIONS_DIR = os.path.join(testapp_with_appconfig.__path__[0], "migrations")


@override_settings(
    MIGRATION_MODULES={
        "testapp_with_appconfig": "tests.testapp_with_appconfig.migrations"
    }
)
class BaseDeleteSquashedMigrationsTestCase(TestCase):
    def migration_exists(self, filename):
        return os.path.exists(os.path.join(MIGRATIONS_DIR, filename))

    def setUp(self):
        class TitleModel(models.Model):
            created_at = models.DateTimeField(auto_now_add=True)
            title = models.CharField(max_length=50)

            class Meta:
                app_label = "testapp_with_appconfig"

        call_command("makemigrations")

    def tearDown(self):
        for root, dirs, files in os.walk(MIGRATIONS_DIR):
            for filename in files:
                if filename.startswith("000"):
                    os.remove(os.path.join(root, filename))


@pytest.mark.xfail
class DeleteSquashedMigrationsExceptionsTests(BaseDeleteSquashedMigrationsTestCase):
    """Tests for delete_squashed_migrations command exceptions."""

    def test_should_raise_CommandError_if_app_does_not_have_migrations(self):
        with self.assertRaisesRegex(
            CommandError,
            r"App 'testapp_with_no_models_file' does not have migrations \(so delete_squashed_migrations on it makes no sense\)",
        ):
            call_command("delete_squashed_migrations", "testapp_with_no_models_file")

    def test_should_raise_CommandEror_if_migration_is_not_squashed(self):
        with self.assertRaisesRegex(
            CommandError,
            "The migration testapp_with_appconfig 0001_initial is not a squashed migration.",
        ):
            call_command("delete_squashed_migrations", "testapp_with_appconfig", "0001")

    def test_should_raise_CommandEror_if_more_than_one_migration_matches_to_given_arg(
        self,
    ):
        class NameModel(models.Model):
            created_at = models.DateTimeField(auto_now_add=True)
            name = models.CharField(max_length=50)

            class Meta:
                app_label = "testapp_with_appconfig"

        call_command("makemigrations", "testapp_with_appconfig")
        call_command("squashmigrations", "testapp_with_appconfig", "0002", "--noinput")

        with self.assertRaisesRegex(
            CommandError,
            "More than one migration matches '0001' in app 'testapp_with_appconfig'. Please be more specific.",
        ):
            call_command("delete_squashed_migrations", "testapp_with_appconfig", "0001")

    def test_should_raise_CommandEror_if_squashed_migration_not_found(self):
        class NameModel(models.Model):
            created_at = models.DateTimeField(auto_now_add=True)
            name = models.CharField(max_length=50)

            class Meta:
                app_label = "testapp_with_appconfig"

        call_command("makemigrations", "testapp_with_appconfig")

        with self.assertRaisesRegex(
            CommandError,
            "Cannot find a squashed migration in app 'testapp_with_appconfig'.",
        ):
            call_command("delete_squashed_migrations", "testapp_with_appconfig")

    def test_should_raise_CommandEror_if_squashed_migration_not_foundee(self):
        with self.assertRaisesRegex(
            CommandError,
            "Cannot find a migration matching '0002' from app 'testapp_with_appconfig'.",
        ):
            call_command("delete_squashed_migrations", "testapp_with_appconfig", "0002")

    def test_should_raise_CommandError_when_database_does_not_exist(self):
        with self.assertRaisesRegex(
            CommandError, "Unknown database non-existing_database"
        ):
            call_command(
                "delete_squashed_migrations", "--database=non-existing_database"
            )


@pytest.mark.xfail
class DeleteSquashedMigrationsTests(BaseDeleteSquashedMigrationsTestCase):
    """Tests for delete_squashed_migrations command."""

    @patch(
        "django_extensions.management.commands.delete_squashed_migrations.six.moves.input"
    )
    def test_should_delete_squashed_migrations(self, m_input):
        m_input.return_value = "y"

        class NameModel(models.Model):
            created_at = models.DateTimeField(auto_now_add=True)
            name = models.CharField(max_length=50)

            class Meta:
                app_label = "testapp_with_appconfig"

        call_command("makemigrations", "testapp_with_appconfig")
        call_command("squashmigrations", "testapp_with_appconfig", "0002", "--noinput")
        call_command("delete_squashed_migrations", "testapp_with_appconfig")

        self.assertFalse(self.migration_exists("0001_initial.py"))
        self.assertFalse(self.migration_exists("0002_namemodel.py"))
        self.assertTrue(self.migration_exists("0001_squashed_0002_namemodel.py"))

    def test_should_delete_squashed_migrations_if_interactive_mode_is_set_to_False(
        self,
    ):
        class NameModel(models.Model):
            created_at = models.DateTimeField(auto_now_add=True)
            name = models.CharField(max_length=50)

            class Meta:
                app_label = "testapp_with_appconfig"

        call_command("makemigrations", "testapp_with_appconfig")
        call_command("squashmigrations", "testapp_with_appconfig", "0002", "--noinput")
        call_command(
            "delete_squashed_migrations", "testapp_with_appconfig", interactive=False
        )

        self.assertFalse(self.migration_exists("0001_initial.py"))
        self.assertFalse(self.migration_exists("0002_namemodel.py"))
        self.assertTrue(self.migration_exists("0001_squashed_0002_namemodel.py"))

    @patch(
        "django_extensions.management.commands.delete_squashed_migrations.six.moves.input"
    )
    def test_should_not_delete_anything(self, m_input):
        m_input.return_value = None

        class NameModel(models.Model):
            created_at = models.DateTimeField(auto_now_add=True)
            name = models.CharField(max_length=50)

            class Meta:
                app_label = "testapp_with_appconfig"

        call_command("makemigrations", "testapp_with_appconfig")
        call_command("squashmigrations", "testapp_with_appconfig", "0002", "--noinput")
        call_command("delete_squashed_migrations", "testapp_with_appconfig")

        self.assertTrue(self.migration_exists("0001_initial.py"))
        self.assertTrue(self.migration_exists("0002_namemodel.py"))
        self.assertTrue(self.migration_exists("0001_squashed_0002_namemodel.py"))

    def test_should_not_delete_files_for_given_squashed_migration(self):
        class NameModel(models.Model):
            created_at = models.DateTimeField(auto_now_add=True)
            name = models.CharField(max_length=50)

            class Meta:
                app_label = "testapp_with_appconfig"

        call_command("makemigrations", "testapp_with_appconfig")
        call_command("squashmigrations", "testapp_with_appconfig", "0002", "--noinput")

        class FooModel(models.Model):
            created_at = models.DateTimeField(auto_now_add=True)
            name = models.CharField(max_length=50)

            class Meta:
                app_label = "testapp_with_appconfig"

        call_command("makemigrations", "testapp_with_appconfig")
        call_command(
            "delete_squashed_migrations",
            "testapp_with_appconfig",
            "0001_squashed_0002_namemodel",
            interactive=False,
        )

        self.assertFalse(self.migration_exists("0001_initial.py"))
        self.assertFalse(self.migration_exists("0002_namemodel.py"))
        self.assertTrue(self.migration_exists("0001_squashed_0002_namemodel.py"))
        self.assertTrue(self.migration_exists("0002_foomodel.py"))



# -*- coding: utf-8 -*-
import os
import shutil

from django.conf import settings
from django.core.management import call_command
from django.test import TestCase
from io import StringIO

from unittest.mock import patch


TEST_APP = "testapp_with_appconfig"


class CreateCommandTests(TestCase):
    """Tests for create_command command."""

    def setUp(self):  # noqa
        self.management_command_path = os.path.join(
            settings.BASE_DIR, "tests/{}/management".format(TEST_APP)
        )
        self.command_template_path = os.path.join(
            settings.BASE_DIR, "django_extensions/conf/command_template"
        )

        self.files = [
            "__init__.py",
            "commands/__init__.py",
            "commands/sample.py",
        ]

    def tearDown(self):  # noqa
        shutil.rmtree(self.management_command_path, ignore_errors=True)
        shutil.rmtree(
            os.path.join(self.command_template_path, ".hidden"), ignore_errors=True
        )
        test_pyc_path = os.path.join(self.command_template_path, "test.pyc")
        if os.path.isfile(test_pyc_path):
            os.remove(test_pyc_path)

    def _create_management_command_with_empty_files(self):
        os.mkdir(self.management_command_path)
        os.mkdir(os.path.join(self.management_command_path, "commands"))
        for f in self.files:
            with open(os.path.join(self.management_command_path, f), "a"):
                pass

    def _create__pycache__in_command_template_directory(self):
        with open(os.path.join(self.command_template_path, "test.pyc"), "a"):
            pass

    def _create_hidden_directory_in_command_template_directory(self):
        os.mkdir(os.path.join(self.command_template_path, ".hidden"))

    @patch("sys.stdout", new_callable=StringIO)
    def test_should_print_management_command_files_only_on_dry_run(self, m_stdout):  # noqa
        call_command("create_command", TEST_APP, "--dry-run", verbosity=2)

        for f in self.files:
            filepath = os.path.join(self.management_command_path, f)
            self.assertIn(filepath, m_stdout.getvalue())
            self.assertFalse(os.path.isfile(filepath))

    @patch("sys.stdout", new_callable=StringIO)
    def test_should_create_management_command_files_and_print_filepaths(self, m_stdout):  # noqa
        call_command("create_command", TEST_APP, verbosity=2)

        for f in self.files:
            filepath = os.path.join(self.management_command_path, f)
            self.assertIn(filepath, m_stdout.getvalue())
            self.assertTrue(os.path.isfile(filepath))

    @patch("sys.stdout", new_callable=StringIO)
    def test_should_print_that_filepaths_already_exists(self, m_stdout):  # noqa
        self._create_management_command_with_empty_files()

        call_command("create_command", TEST_APP, verbosity=2)

        for f in self.files:
            filepath = os.path.join(self.management_command_path, f)
            self.assertIn("{} already exists".format(filepath), m_stdout.getvalue())
            self.assertTrue(os.path.isfile(filepath))
            self.assertEqual(os.path.getsize(filepath), 0)

    @patch("sys.stderr", new_callable=StringIO)
    @patch("django_extensions.management.commands.create_command._make_writeable")  # noqa
    def test_should_print_error_on_OSError_exception(self, m__make_writeable, m_stderr):  # noqa
        m__make_writeable.side_effect = OSError
        self._create__pycache__in_command_template_directory()
        self._create_hidden_directory_in_command_template_directory()

        call_command("create_command", TEST_APP)
        for f in self.files:
            filepath = os.path.join(self.management_command_path, f)
            self.assertIn(
                "Notice: Couldn't set permission bits on {}. You're probably using an uncommon filesystem setup. No problem.\n".format(
                    filepath
                ),  # noqa
                m_stderr.getvalue(),
            )



# -*- coding: utf-8 -*-
import fnmatch
import os

from io import StringIO
from django.core.management import call_command
from django.test import TestCase


class CompilePycTests(TestCase):
    def setUp(self):
        self.project_root = os.path.join("tests", "testapp")
        self._settings = os.environ.get("DJANGO_SETTINGS_MODULE")
        os.environ["DJANGO_SETTINGS_MODULE"] = "django_extensions.settings"

    def tearDown(self):
        if self._settings:
            os.environ["DJANGO_SETTINGS_MODULE"] = self._settings

    def _find_pyc(self, path, mask="*.pyc"):
        pyc_glob = []
        for root, dirs, filenames in os.walk(path):
            for filename in fnmatch.filter(filenames, mask):
                pyc_glob.append(os.path.join(root, filename))
        return pyc_glob

    def test_compiles_pyc_files(self):
        with self.settings(BASE_DIR=self.project_root):
            call_command("clean_pyc")
        pyc_glob = self._find_pyc(self.project_root)
        self.assertEqual(len(pyc_glob), 0)
        with self.settings(BASE_DIR=self.project_root):
            call_command("compile_pyc")
        pyc_glob = self._find_pyc(self.project_root)
        self.assertTrue(len(pyc_glob) > 0)
        with self.settings(BASE_DIR=self.project_root):
            call_command("clean_pyc")

    def test_takes_path(self):
        out = StringIO()
        with self.settings(BASE_DIR=""):
            call_command("clean_pyc", path=self.project_root)
        pyc_glob = self._find_pyc(self.project_root)
        self.assertEqual(len(pyc_glob), 0)
        with self.settings(BASE_DIR=""):
            call_command("compile_pyc", verbosity=2, path=self.project_root, stdout=out)
        expected = [
            "Compiling %s..." % fn
            for fn in sorted(self._find_pyc(self.project_root, mask="*.py"))
        ]
        output = out.getvalue().splitlines()
        self.assertEqual(expected, sorted(output))
        with self.settings(BASE_DIR=""):
            call_command("clean_pyc", path=self.project_root)



# -*- coding: utf-8 -*-
from django.test import SimpleTestCase
from django_extensions.management import color
from . import force_color_support


class ColorTest(SimpleTestCase):
    def test_no_style(self):
        with force_color_support:
            style = color.no_style().MODULE_NAME
            text = "csv"
            styled_text = style(text)
            self.assertEqual(text, styled_text)

    def test_color_style(self):
        with force_color_support:
            style = color.color_style().MODULE_NAME
            text = "antigravity"
            styled_text = style(text)
            self.assertIn(text, styled_text)
            self.assertNotEqual(text, styled_text)



# -*- coding: utf-8 -*-
import importlib

from django.conf import settings
from django.contrib.auth import load_backend, BACKEND_SESSION_KEY, SESSION_KEY
from django.contrib.sessions.backends.base import VALID_KEY_CHARS
from django.core.management.base import BaseCommand, CommandError
from django_extensions.management.utils import signalcommand


class Command(BaseCommand):
    help = (
        "print the user information for the provided session key. "
        "this is very helpful when trying to track down the person who "
        "experienced a site crash."
    )

    def add_arguments(self, parser):
        parser.add_argument("session_id", nargs="+", type=str, help="user session id")

    @signalcommand
    def handle(self, *args, **options):
        key = options["session_id"][0]

        if not set(key).issubset(set(VALID_KEY_CHARS)):
            raise CommandError("malformed session key")

        engine = importlib.import_module(settings.SESSION_ENGINE)

        if not engine.SessionStore().exists(key):
            print("Session Key does not exist. Expired?")
            return

        session = engine.SessionStore(key)
        data = session.load()

        print("Session to Expire: %s" % session.get_expiry_date())
        print("Raw Data: %s" % data)
        uid = data.get(SESSION_KEY, None)
        backend_path = data.get(BACKEND_SESSION_KEY, None)

        if backend_path is None:
            print("No authentication backend associated with session")
            return

        if uid is None:
            print("No user associated with session")
            return

        print("User id: %s" % uid)

        backend = load_backend(backend_path)
        user = backend.get_user(user_id=uid)
        if user is None:
            print("No user associated with that id.")
            return

        # use django standrd api for reporting
        print("full name: %s" % user.get_full_name())
        print("short name: %s" % user.get_short_name())
        print("username: %s" % user.get_username())
        if hasattr(user, "email"):
            print("email: %s" % user.email)



# -*- coding: utf-8 -*-
from django_extensions.management.jobs import YearlyJob

from unittest import mock


YEARLY_JOB_MOCK = mock.MagicMock()


class Job(YearlyJob):
    help = "My sample yearly job."

    def execute(self):
        YEARLY_JOB_MOCK()



# -*- coding: utf-8 -*-
from datetime import datetime

from django.test import TestCase

from .testapp.models import (
    ModelModificationDateTimeField,
    CustomModelModificationDateTimeField,
    DisabledUpdateModelModificationDateTimeField,
)


class ModificationDatetimeFieldTest(TestCase):
    def test_update_model_with_modification_field(self):
        m = ModelModificationDateTimeField.objects.create()
        current_updated_time = m.modified
        m.field_to_update = False
        m.save()
        self.assertNotEqual(m.modified, current_updated_time)

    def test_custom_modification_field_name(self):
        m = CustomModelModificationDateTimeField.objects.create()
        current_updated_time = m.custom_modified
        m.field_to_update = False
        m.save()
        self.assertNotEqual(m.custom_modified, current_updated_time)

    def test_disabled_update_modification_field(self):
        m = DisabledUpdateModelModificationDateTimeField.objects.create(
            modified=datetime.now()
        )
        current_updated_time = m.modified
        m.field_to_update = False
        m.save()
        self.assertEqual(m.modified, current_updated_time)



# -*- coding: utf-8 -*-
from unittest.mock import patch

from django.core.exceptions import ValidationError
from django.test import TestCase, SimpleTestCase

from django_extensions.validators import (
    NoControlCharactersValidator,
    NoWhitespaceValidator,
    HexValidator,
)


class NoControlCharactersValidatorTests(TestCase):
    """Tests for NoControlCharactersValidator."""

    def test_should_raise_default_message_and_code_if_value_contains_new_line(self):
        self.validator = NoControlCharactersValidator()
        value_with_new_line = "test\nvalue"

        with self.assertRaises(ValidationError) as cm:
            self.validator(value_with_new_line)

        self.assertEqual(
            cm.exception.message,
            "Control Characters like new lines or tabs are not allowed.",
        )
        self.assertEqual(cm.exception.code, "no_control_characters")
        self.assertDictEqual(
            cm.exception.params, {"value": value_with_new_line, "whitelist": None}
        )

    def test_should_raise_custom_message_and_code_if_value_contains_tabs(self):
        self.validator = NoControlCharactersValidator(
            message="custom message", code="custom code"
        )
        value_with_tabs = "test\tvalue"

        with self.assertRaises(ValidationError) as cm:
            self.validator(value_with_tabs)

        self.assertEqual(cm.exception.message, "custom message")
        self.assertEqual(cm.exception.code, "custom code")
        self.assertDictEqual(
            cm.exception.params, {"value": value_with_tabs, "whitelist": None}
        )

    def test_should_not_raise_if_value_contains_characters_which_is_on_whitelist(self):
        self.validator = NoControlCharactersValidator(
            message="custom message", code="custom code", whitelist=["\n"]
        )
        value_with_new_line = "test\nvalue"

        result = self.validator(value_with_new_line)

        self.assertIsNone(result)


class NoWhiteSpaceValidatorTests(TestCase):
    """Tests for NoWhitespaceValidator."""

    def test_should_raise_default_message_and_code_if_value_has_leading_whitespace(
        self,
    ):
        self.validator = NoWhitespaceValidator()

        value_with_leading_whitespace = " test_value"

        with self.assertRaises(ValidationError) as cm:
            self.validator(value_with_leading_whitespace)

        self.assertEqual(
            cm.exception.message, "Leading and Trailing whitespaces are not allowed."
        )
        self.assertEqual(cm.exception.code, "no_whitespace")
        self.assertDictEqual(
            cm.exception.params, {"value": value_with_leading_whitespace}
        )

    def test_should_raise_custom_message_and_code_if_value_has_trailing_whitespace(
        self,
    ):
        self.validator = NoWhitespaceValidator(
            message="custom message", code="custom code"
        )
        value_with_trailing_whitespace = "test value "

        with self.assertRaises(ValidationError) as cm:
            self.validator(value_with_trailing_whitespace)

        self.assertEqual(cm.exception.message, "custom message")
        self.assertEqual(cm.exception.code, "custom code")
        self.assertDictEqual(
            cm.exception.params, {"value": value_with_trailing_whitespace}
        )

    def test_should_not_raise_if_value_doesnt_have_leading_or_trailing_whitespaces(
        self,
    ):
        self.validator = NoWhitespaceValidator()
        value_without_leading_or_trailing_whitespaces = "test value"

        result = self.validator(value_without_leading_or_trailing_whitespaces)

        self.assertIsNone(result)


class TestHexValidator(SimpleTestCase):
    def test_custom_message_and_code(self):
        self.validator = HexValidator(message="message", code="code")

        self.assertEqual(self.validator.message, "message")
        self.assertEqual(self.validator.code, "code")

    def test_equality_of_objs_with_obj_of_different_type(self):
        self.assertNotEqual(TypeError(), HexValidator())

    def test_equality_of_objs_with_different_code(self):
        self.assertNotEqual(HexValidator(code="1"), HexValidator(code="a"))

    def test_equality_of_objs_with_different_message(self):
        self.assertNotEqual(
            HexValidator(code="code", message="a"),
            HexValidator(code="code", message="acb"),
        )

    def test_equality_of_objs_with_same_code_and_message(self):
        self.assertEqual(
            HexValidator(code="c", message="m"), HexValidator(code="c", message="m")
        )

    def test_fixed_length(self):
        value = "abcd"
        self.validator = HexValidator(length=5)

        with self.assertRaises(ValidationError) as err:
            self.validator(value)

        self.assertEqual(
            str(err.exception), "['Invalid length. Must be 5 characters.']"
        )
        self.assertEqual(err.exception.code, "hex_only_length")

    def test_min_length(self):
        value = "a"
        self.validator = HexValidator(min_length=5)

        with self.assertRaises(ValidationError) as err:
            self.validator(value)

        self.assertEqual(
            str(err.exception), "['Ensure that there are more than 5 characters.']"
        )
        self.assertEqual(err.exception.code, "hex_only_min_length")

    def test_with_max_length(self):
        value = "abcd"
        self.validator = HexValidator(max_length=2)

        with self.assertRaises(ValidationError) as err:
            self.validator(value)

        self.assertEqual(
            str(err.exception), "['Ensure that there are no more than 2 characters.']"
        )
        self.assertEqual(err.exception.code, "hex_only_max_length")

    def test_invalid_type(self):
        value = 1
        with patch("django_extensions.validators.force_str", return_value=1):
            self.validator = HexValidator()

            with self.assertRaises(ValidationError) as err:
                self.validator(value)

        self.assertEqual(str(err.exception), "['Only a hex string is allowed.']")
        self.assertEqual(err.exception.code, "hex_only")

    def test_invalid_hex(self):
        value = "1"
        self.validator = HexValidator()

        with self.assertRaises(ValidationError) as err:
            self.validator(value)

        self.assertEqual(str(err.exception), "['Only a hex string is allowed.']")
        self.assertEqual(err.exception.code, "hex_only")

    def test_valid_hex(self):
        value = "b901ef"
        self.validator = HexValidator()

        result = self.validator(value)

        self.assertIsNone(result)



# -*- coding: utf-8 -*-
import sys
from io import StringIO
from typing import Dict, Set, Type  # noqa

from django.test import TestCase
from django_extensions.management.commands import shell_plus


class AutomaticShellPlusImportsTestCase(TestCase):
    def setUp(self):
        super().setUp()
        sys.stdout = StringIO()
        sys.stderr = StringIO()
        self.imported_objects = {}  # type: Dict[str, Type]
        self.output = ""

    def get_all_names_for_class(self, model_to_find_occurrences):  # type: (Type) -> Set[str]
        """
        Returns all names under current class is imported.
        :param model_to_find_occurrences: class to find names
        :return: set of names under class is imported.
        """
        result = set()
        for name, model_class in self.imported_objects.items():
            if model_class == model_to_find_occurrences:
                result.add(name)
        return result

    def assert_imported_under_names(self, model_class, names_under_model_is_available):  # type: (Type, Set[str]) -> ()
        """
        Function which asserts that class is available under given names and not available under any other name.
        :param model_class: class to assert availability.
        :param names_under_model_is_available: names under which class should be available.
        """
        self.assertSetEqual(
            self.get_all_names_for_class(model_class), names_under_model_is_available
        )
        imports_output = self.output.split("from ")
        for line in imports_output:
            if line.startswith(model_class.__module__):
                for name in names_under_model_is_available:
                    # assert that in print imports this model occurs only under names from parameter
                    if name == model_class.__name__:
                        expected_output = name
                    else:
                        expected_output = "%s (as %s)" % (model_class.__name__, name)
                    line = line.replace(expected_output, "", 1)
                self.assertNotIn(line, model_class.__name__)

    def run_shell_plus(self):
        command = shell_plus.Command()
        self.imported_objects = command.get_imported_objects({})
        self.output = sys.stdout.getvalue()



# -*- coding: utf-8 -*-
from unittest.mock import Mock

from django.test import RequestFactory, TestCase
from factory import Iterator

from django_extensions.admin.filter import NotNullFieldListFilter, NullFieldListFilter

from .testapp.factories import SecretFactory
from .testapp.models import Secret


class BaseFieldFilter(TestCase):
    """Base class for filter test cases."""

    @classmethod
    def setUpClass(cls):
        SecretFactory.create_batch(5, text=Iterator([None, None, "foo", "bar", None]))
        cls.request = RequestFactory().get("/admin/testapp/secret")
        cls.field = Secret._meta.get_field("text")
        cls.field_path = "text"
        cls.qs = Secret.objects.all()

    @classmethod
    def tearDownClass(cls):
        Secret.objects.all().delete()


class NullFieldListFilterTests(BaseFieldFilter):
    """Tests for NullFieldListFilter."""

    def test_should_not_filter_qs_if_all_lookup_selected(self):
        params = {}
        filter_spec = NullFieldListFilter(
            self.field, self.request, params, Secret, Mock(), self.field_path
        )

        result = filter_spec.queryset(self.request, self.qs)

        self.assertCountEqual(self.qs, result)

    def test_should_return_objects_with_empty_text_if_yes_lookup_selected(self):
        expected_result = Secret.objects.filter(text__isnull=True)
        params = {"text__isnull": "1"}
        filter_spec = NullFieldListFilter(
            self.field, self.request, params, Secret, Mock(), self.field_path
        )

        result = filter_spec.queryset(self.request, self.qs)

        self.assertCountEqual(expected_result, result)

    def test_should_return_objects_with_not_empty_text_value_if_no_lookup_selected(
        self,
    ):
        expected_result = Secret.objects.filter(text__isnull=False)
        params = {"text__isnull": "0"}
        filter_spec = NullFieldListFilter(
            self.field, self.request, params, Secret, Mock(), self.field_path
        )

        result = filter_spec.queryset(self.request, self.qs)

        self.assertCountEqual(expected_result, result)

    def test_choices(self):
        expected_result = [
            {"selected": True, "query_string": "?", "display": "All"},
            {"selected": False, "query_string": "?active__isnull=1", "display": "Yes"},
            {"selected": False, "query_string": "?active__isnull=0", "display": "No"},
        ]
        m_cl = Mock()
        m_cl.get_query_string.side_effect = [
            "?",
            "?active__isnull=1",
            "?active__isnull=0",
        ]
        filter_spec = NullFieldListFilter(
            self.field, self.request, {}, Secret, Mock(), self.field_path
        )

        result = filter_spec.choices(m_cl)

        self.assertEqual(list(result), expected_result)


class NotNullFieldListFilterTests(BaseFieldFilter):
    """Tests for NotNullFieldListFilter."""

    def test_should_not_filter_qs_if_all_lookup_selected(self):
        params = {}
        filter_spec = NotNullFieldListFilter(
            self.field, self.request, params, Secret, Mock(), self.field_path
        )

        result = filter_spec.queryset(self.request, self.qs)

        self.assertCountEqual(self.qs, result)



# -*- coding: utf-8 -*-
import pytest

from django.db import migrations, models
from django.db.migrations.writer import MigrationWriter
from django.test import TestCase
from django.utils.encoding import force_bytes

import django_extensions  # noqa
from django_extensions.db.fields import AutoSlugField

from .testapp.models import (
    ChildSluggedTestModel,
    CustomFuncPrecedenceSluggedTestModel,
    CustomFuncSluggedTestModel,
    FKSluggedTestModel,
    FKSluggedTestModelCallable,
    FunctionSluggedTestModel,
    ModelMethodSluggedTestModel,
    SluggedTestModel,
    SluggedTestNoOverwriteOnAddModel,
    OverridedFindUniqueModel,
    SluggedWithConstraintsTestModel,
    SluggedWithUniqueTogetherTestModel,
)


@pytest.mark.usefixtures("admin_user")
class AutoSlugFieldTest(TestCase):
    def tearDown(self):
        super().tearDown()

        SluggedTestModel.objects.all().delete()
        CustomFuncSluggedTestModel.objects.all().delete()
        CustomFuncPrecedenceSluggedTestModel.objects.all().delete()

    def test_auto_create_slug(self):
        m = SluggedTestModel(title="foo")
        m.save()
        self.assertEqual(m.slug, "foo")

    def test_auto_create_next_slug(self):
        m = SluggedTestModel(title="foo")
        m.save()

        m = SluggedTestModel(title="foo")
        m.save()
        self.assertEqual(m.slug, "foo-2")

    def test_auto_create_slug_with_number(self):
        m = SluggedTestModel(title="foo 2012")
        m.save()
        self.assertEqual(m.slug, "foo-2012")

    def test_auto_update_slug_with_number(self):
        m = SluggedTestModel(title="foo 2012")
        m.save()
        m.save()
        self.assertEqual(m.slug, "foo-2012")

    def test_update_slug(self):
        m = SluggedTestModel(title="foo")
        m.save()
        self.assertEqual(m.slug, "foo")

        # update m instance without using `save'
        SluggedTestModel.objects.filter(pk=m.pk).update(slug="foo-2012")
        # update m instance with new data from the db
        m = SluggedTestModel.objects.get(pk=m.pk)
        self.assertEqual(m.slug, "foo-2012")

        m.save()
        self.assertEqual(m.title, "foo")
        self.assertEqual(m.slug, "foo-2012")

        # Check slug is not overwrite
        m.title = "bar"
        m.save()
        self.assertEqual(m.title, "bar")
        self.assertEqual(m.slug, "foo-2012")

    def test_simple_slug_source(self):
        m = SluggedTestModel(title="-foo")
        m.save()
        self.assertEqual(m.slug, "foo")

        n = SluggedTestModel(title="-foo")
        n.save()
        self.assertEqual(n.slug, "foo-2")

        n.save()
        self.assertEqual(n.slug, "foo-2")

    def test_empty_slug_source(self):
        # regression test

        m = SluggedTestModel(title="")
        m.save()
        self.assertEqual(m.slug, "-2")

        n = SluggedTestModel(title="")
        n.save()
        self.assertEqual(n.slug, "-3")

        n.save()
        self.assertEqual(n.slug, "-3")

    def test_callable_method_slug_source(self):
        m = ModelMethodSluggedTestModel(title="-foo")
        m.save()
        self.assertEqual(m.slug, "the-title-is-foo")

        n = ModelMethodSluggedTestModel(title="-foo")
        n.save()
        self.assertEqual(n.slug, "the-title-is-foo-2")

        n.save()
        self.assertEqual(n.slug, "the-title-is-foo-2")

    def test_callable_function_slug_source(self):
        m = FunctionSluggedTestModel(title="-foo")
        m.save()
        self.assertEqual(m.slug, "the-title-is-foo")

        n = FunctionSluggedTestModel(title="-foo")
        n.save()
        self.assertEqual(n.slug, "the-title-is-foo-2")

        n.save()
        self.assertEqual(n.slug, "the-title-is-foo-2")

    def test_inheritance_creates_next_slug(self):
        m = SluggedTestModel(title="foo")
        m.save()

        n = ChildSluggedTestModel(title="foo")
        n.save()
        self.assertEqual(n.slug, "foo-2")

        o = SluggedTestModel(title="foo")
        o.save()
        self.assertEqual(o.slug, "foo-3")

    def test_foreign_key_populate_from_field(self):
        m_fk = SluggedTestModel(title="foo")
        m_fk.save()
        m = FKSluggedTestModel(related_field=m_fk)
        m.save()
        self.assertEqual(m.slug, "foo")

    def test_foreign_key_populate_from_callable(self):
        m_fk = ModelMethodSluggedTestModel(title="foo")
        m_fk.save()
        m = FKSluggedTestModelCallable(related_field=m_fk)
        m.save()
        self.assertEqual(m.slug, "the-title-is-foo")

    def test_copy_model_generates_new_slug(self):
        m = SluggedTestModel(title="foo")
        m.save()
        self.assertEqual(m.slug, "foo")

        m.pk = None
        m.save()
        self.assertEqual(m.slug, "foo-2")

    def test_copy_model_generates_new_slug_no_overwrite_on_add(self):
        m = SluggedTestNoOverwriteOnAddModel(title="foo")
        m.save()
        self.assertEqual(m.slug, "foo")

        m.pk = None
        m.slug = None
        m.save()
        self.assertEqual(m.slug, "foo-2")

    def test_populate_from_does_not_allow_bytes(self):
        with pytest.raises(TypeError):
            AutoSlugField(populate_from=b"bytes")

        with pytest.raises(TypeError):
            AutoSlugField(populate_from=[b"bytes"])

    def test_populate_from_must_allow_string_or_list_str_or_tuple_str(self):
        AutoSlugField(populate_from="str")
        AutoSlugField(populate_from=["str"])
        AutoSlugField(populate_from=("str",))

    def test_slug_argument_priority(self):
        m = SluggedTestModel(slug="slug", title="title")
        m.save()
        self.assertEqual(m.slug, "title")

    def test_slug_argument_priority_no_overwrite_on_add(self):
        m = SluggedTestNoOverwriteOnAddModel(slug="slug", title="title")
        m.save()
        self.assertEqual(m.slug, "slug")

    def test_overrided_find_unique_autoslug_field(self):
        m = OverridedFindUniqueModel(title="foo")
        slug_field = m._meta.fields[2]
        self.assertFalse(hasattr(slug_field, "overrided"))
        m.save()
        slug_field = m._meta.fields[2]
        self.assertTrue(slug_field.overrided)

    def test_slugify_func(self):
        to_upper = lambda c: c.upper()
        to_lower = lambda c: c.lower()

        content_n_func_n_expected = (
            ("test", to_upper, "TEST"),
            ("", to_upper, ""),
            ("TEST", to_lower, "test"),
        )

        for content, slugify_function, expected in content_n_func_n_expected:
            self.assertEqual(
                AutoSlugField.slugify_func(content, slugify_function), expected
            )

    def test_use_custom_slug_function(self):
        m = CustomFuncSluggedTestModel(title="test")
        m.save()
        self.assertEqual(m.slug, "TEST")

    def test_precedence_custom_slug_function(self):
        m = CustomFuncPrecedenceSluggedTestModel(title="test")
        m.save()
        self.assertEqual(m.slug, "TEST")
        self.assertTrue(hasattr(m._meta.get_field("slug"), "slugify_function"))
        self.assertEqual(m._meta.get_field("slug").slugify_function("TEST"), "test")

    def test_auto_create_slug_with_unique_together(self):
        m = SluggedWithUniqueTogetherTestModel(
            title="foo", category="self-introduction"
        )
        m.save()
        self.assertEqual(m.slug, "foo")

        m = SluggedWithUniqueTogetherTestModel(title="foo", category="review")
        m.save()
        self.assertEqual(m.slug, "foo")

        # check if satisfy database integrity
        m = SluggedWithUniqueTogetherTestModel(title="foo", category="review")
        m.save()
        self.assertEqual(m.slug, "foo-2")

    def test_auto_create_slug_with_constraints(self):
        m = SluggedWithConstraintsTestModel(title="foo", category="self-introduction")
        m.save()
        self.assertEqual(m.slug, "foo")

        m = SluggedWithConstraintsTestModel(title="foo", category="review")
        m.save()
        self.assertEqual(m.slug, "foo")

        # check if satisfy database integrity
        m = SluggedWithConstraintsTestModel(title="foo", category="review")
        m.save()
        self.assertEqual(m.slug, "foo-2")


class MigrationTest(TestCase):
    def safe_exec(self, string, value=None):
        dct = {}
        try:
            exec(force_bytes(string), globals(), dct)
        except Exception as e:
            if value:
                self.fail(
                    "Could not exec %r (from value %r): %s" % (string.strip(), value, e)
                )
            else:
                self.fail("Could not exec %r: %s" % (string.strip(), e))
        return dct

    def test_17_migration(self):
        """
        Tests making migrations with Django's migration framework
        """

        fields = {
            "autoslugfield": AutoSlugField(populate_from="otherfield"),
        }

        migration = type(
            str("Migration"),
            (migrations.Migration,),
            {
                "operations": [
                    migrations.CreateModel(
                        "MyModel",
                        tuple(fields.items()),
                        {"populate_from": "otherfield"},
                        (models.Model,),
                    ),
                ],
            },
        )
        writer = MigrationWriter(migration)
        output = writer.as_string()

        self.assertIsInstance(output, str, "Migration as_string returned bytes")

        # We don't test the output formatting - that's too fragile.
        # Just make sure it runs for now, and that things look alright.
        result = self.safe_exec(output)
        self.assertIn("Migration", result)

    def test_stable_deconstruct(self):
        slug_field = SluggedTestModel._meta.get_field("slug")
        construction_values = slug_field.deconstruct()
        m = SluggedTestModel(title="foo")
        m.save()
        self.assertEqual(slug_field.deconstruct(), construction_values)



# -*- coding: utf-8 -*-
from django.test import TestCase, RequestFactory
from django.http import HttpResponse
from django.views.generic import DetailView
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import PermissionDenied

from django_extensions.auth.mixins import ModelUserFieldPermissionMixin

from tests.testapp.models import HasOwnerModel


class EmptyResponseView(DetailView):
    model = HasOwnerModel

    def get(self, request, *args, **kwargs):
        return HttpResponse()


class OwnerView(ModelUserFieldPermissionMixin, EmptyResponseView):
    model_permission_user_field = "owner"


class ModelUserFieldPermissionMixinTests(TestCase):
    factory = RequestFactory()
    User = get_user_model()

    @classmethod
    def setUpTestData(cls):
        cls.user = cls.User.objects.create(username="Joe", password="pass")
        cls.ownerModel = HasOwnerModel.objects.create(owner=cls.user)

    # Test if owner model has access
    def test_permission_pass(self):
        request = self.factory.get("/permission-required/" + str(self.ownerModel.id))
        request.user = self.user
        resp = OwnerView.as_view()(request)
        self.assertEqual(resp.status_code, 200)

    # # Test if non owner model is redirected
    def test_permission_denied_and_redirect(self):
        request = self.factory.get("/permission-required/" + str(self.ownerModel.id))
        request.user = AnonymousUser()
        resp = OwnerView.as_view()(request)
        self.assertRaises(PermissionDenied)
        self.assertEqual(resp.status_code, 302)



# -*- coding: utf-8 -*-
import os

import pytest
from django.contrib.auth.models import User
from django.core.management import call_command, CommandError
from django.test import TestCase
from django.test.utils import override_settings
from io import StringIO

from unittest.mock import patch


TEST_FIXTURE_DIR = os.path.join(os.path.dirname(__file__), "fixtures")


@override_settings(FIXTURE_DIRS=[TEST_FIXTURE_DIR])
class SyncDataExceptionsTests(TestCase):
    """Tests for SyncData command exceptions."""

    def test_should_return_SyncDataError_when_unknown_fixture_format(self):
        with pytest.raises(
            CommandError,
            match="Problem installing fixture 'foo': jpeg is not a known serialization format.",
        ):
            call_command("syncdata", "foo.jpeg", verbosity=2)

    def test_should_return_SyncDataError_when_file_not_contains_valid_fixture_data(
        self,
    ):
        with pytest.raises(
            CommandError,
            match=r"No fixture data found for 'invalid_fixture'. \(File format may be invalid.\)",
        ):
            call_command("syncdata", "invalid_fixture.xml", verbosity=2)

    def test_should_return_SyncDataError_when_file_has_non_existent_field_in_fixture_data(
        self,
    ):
        with pytest.raises(
            CommandError,
            match=r"Problem installing fixture '.+fixture_with_nonexistent_field.json'",
        ):
            call_command("syncdata", "fixture_with_nonexistent_field.json", verbosity=1)
        with pytest.raises(
            CommandError,
            match="django.core.exceptions.FieldDoesNotExist: User has no field named 'non_existent_field'",
        ):
            call_command("syncdata", "fixture_with_nonexistent_field.json", verbosity=1)

    def test_should_return_SyncDataError_when_multiple_fixtures(self):
        with pytest.raises(
            CommandError,
            match="Multiple fixtures named 'users' in '{}'. Aborting.".format(
                TEST_FIXTURE_DIR
            ),
        ):
            call_command("syncdata", "users", verbosity=2)


@override_settings(FIXTURE_DIRS=[TEST_FIXTURE_DIR])
class SyncDataTests(TestCase):
    """Tests for syncdata command."""

    @patch("sys.stdout", new_callable=StringIO)
    def test_should_print_No_fixtures_found_if_fixture_labels_not_provided(
        self, m_stdout
    ):
        call_command("syncdata", verbosity=2)

        self.assertEqual("No fixtures found.\n", m_stdout.getvalue())

    @patch("sys.stdout", new_callable=StringIO)
    def test_should_print_No_fixtures_found_if_fixtures_not_found(self, m_stdout):
        call_command("syncdata", "foo", verbosity=2)

        self.assertIn("No fixtures found.\n", m_stdout.getvalue())

    def test_should_keep_old_objects_and_load_data_from_json_fixture(self):
        User.objects.all().delete()
        User.objects.create(username="foo")

        call_command(
            "syncdata",
            "--skip-remove",
            os.path.join(TEST_FIXTURE_DIR, "users.json"),
            verbosity=2,
        )

        self.assertTrue(User.objects.filter(username="jdoe").exists())
        self.assertTrue(User.objects.filter(username="foo").exists())

    @patch("sys.stdout", new_callable=StringIO)
    def test_should_delete_old_objects_and_load_data_from_json_fixture(self, m_stdout):
        User.objects.all().delete()
        User.objects.create(username="foo")

        call_command("syncdata", "users.json", verbosity=2)

        self.assertTrue(User.objects.filter(username="jdoe").exists())
        self.assertEqual(User.objects.count(), 1)
        self.assertIn("Installed 1 object from 1 fixture", m_stdout.getvalue())



# -*- coding: utf-8 -*-
from django.core.management import call_command
from django.test import TestCase
from io import StringIO

from unittest.mock import patch


class FindTemplateTests(TestCase):
    @patch("sys.stdout", new_callable=StringIO)
    def test_finding_template(self, m_stdout):
        call_command("find_template", "admin/change_form.html")

        self.assertIn("admin/change_form.html", m_stdout.getvalue())

    @patch("sys.stderr", new_callable=StringIO)
    def test_should_print_error_when_template_not_found(self, m_stderr):
        call_command("find_template", "not_found_template.html")

        self.assertIn("No template found", m_stderr.getvalue())



# -*- coding: utf-8 -*-
from django.core.management import call_command
from unittest import mock


def test_initialize_mail_server():
    with mock.patch(
        "django_extensions.management.commands.mail_debug.asyncio"
    ) as asyncio:
        call_command("mail_debug", "2525")
        assert asyncio.get_event_loop.called, "asyncio.get_event_loop was not called"



# -*- coding: utf-8 -*-
import os
from unittest import mock
import logging
import importlib

from django.core.management import (
    call_command,
    find_commands,
    load_command_class,
    BaseCommand,
)
from django.test import TestCase
from io import StringIO

from django_extensions.management.modelviz import use_model, generate_graph_data
from django_extensions.management.commands.merge_model_instances import (
    get_model_to_deduplicate,
    get_field_names,
    keep_first_or_last_instance,
)
from django_extensions.management.utils import signalcommand
from . import force_color_support
from .testapp.models import (
    Person,
    Name,
    Note,
    Personality,
    Club,
    Membership,
    Permission,
)
from .testapp.jobs.hourly.test_hourly_job import HOURLY_JOB_MOCK
from .testapp.jobs.daily.test_daily_job import DAILY_JOB_MOCK
from .testapp.jobs.weekly.test_weekly_job import WEEKLY_JOB_MOCK
from .testapp.jobs.monthly.test_monthly_job import MONTHLY_JOB_MOCK
from .testapp.jobs.yearly.test_yearly_job import YEARLY_JOB_MOCK


class MockLoggingHandler(logging.Handler):
    """Mock logging handler to check for expected logs."""

    def __init__(self, *args, **kwargs):
        self.reset()
        logging.Handler.__init__(self, *args, **kwargs)

    def emit(self, record):
        self.messages[record.levelname.lower()].append(record.getMessage())

    def reset(self):
        self.messages = {
            "debug": [],
            "info": [],
            "warning": [],
            "error": [],
            "critical": [],
        }


class CommandTest(TestCase):
    def test_error_logging(self):
        # Ensure command errors are properly logged and reraised
        from django_extensions.management.base import logger

        logger.addHandler(MockLoggingHandler())
        module_path = "tests.management.commands.error_raising_command"
        module = importlib.import_module(module_path)
        error_raising_command = module.Command()
        self.assertRaises(Exception, error_raising_command.execute)
        handler = logger.handlers[0]
        self.assertEqual(len(handler.messages["error"]), 1)


class ShowTemplateTagsTests(TestCase):
    def test_some_output(self):
        out = StringIO()
        call_command("show_template_tags", stdout=out)
        output = out.getvalue()
        # Once django_extension is installed during tests it should appear with
        # its templatetags
        self.assertIn("django_extensions", output)
        # let's check at least one
        self.assertIn("syntax_color", output)


class DescribeFormTests(TestCase):
    def test_command(self):
        out = StringIO()
        call_command("describe_form", "django_extensions.Secret", stdout=out)
        output = out.getvalue()
        self.assertIn("class SecretForm(forms.Form):", output)
        self.assertRegex(output, r"name = forms.CharField\(.*max_length=255")
        self.assertRegex(output, r"name = forms.CharField\(.*required=False")
        self.assertRegex(output, r"name = forms.CharField\(.*label=u?'Name'")
        self.assertRegex(output, r"text = forms.CharField\(.*required=False")
        self.assertRegex(output, r"text = forms.CharField\(.*label=u?'Text'")


class CommandSignalTests(TestCase):
    pre = None
    post = None

    def test_works(self):
        from django_extensions.management.signals import post_command, pre_command
        from django_extensions.management.commands.show_template_tags import Command

        def pre(sender, **kwargs):
            CommandSignalTests.pre = dict(**kwargs)

        def post(sender, **kwargs):
            CommandSignalTests.post = dict(**kwargs)

        pre_command.connect(pre, Command)
        post_command.connect(post, Command)

        out = StringIO()
        call_command("show_template_tags", stdout=out)

        self.assertIn("args", CommandSignalTests.pre)
        self.assertIn("kwargs", CommandSignalTests.pre)

        self.assertIn("args", CommandSignalTests.post)
        self.assertIn("kwargs", CommandSignalTests.post)
        self.assertIn("outcome", CommandSignalTests.post)

    def test_exception(self):
        from django_extensions.management.signals import post_command, pre_command

        exception = Exception()

        class TestCommand(BaseCommand):
            @signalcommand
            def handle(self, *args, **options):
                raise exception

        def pre(sender, **kwargs):
            CommandSignalTests.pre = dict(**kwargs)

        def post(sender, **kwargs):
            CommandSignalTests.post = dict(**kwargs)

        pre_command.connect(pre, TestCommand)
        post_command.connect(post, TestCommand)

        out = StringIO()

        with self.assertRaises(Exception) as e_info:
            call_command(TestCommand(), stdout=out)
        self.assertEqual(exception, e_info.exception)
        self.assertIn("args", CommandSignalTests.pre)
        self.assertIn("kwargs", CommandSignalTests.pre)

        self.assertIn("args", CommandSignalTests.post)
        self.assertIn("kwargs", CommandSignalTests.post)
        self.assertIn("outcome", CommandSignalTests.post)
        self.assertEqual(e_info.exception, CommandSignalTests.post["outcome"])


class CommandClassTests(TestCase):
    def setUp(self):
        management_dir = os.path.join("django_extensions", "management")
        self.commands = find_commands(management_dir)

    def test_load_commands(self):
        """Try to load every management command to catch exceptions."""
        try:
            for command in self.commands:
                load_command_class("django_extensions", command)
        except Exception as e:
            self.fail("Can't load command class of {0}\n{1}".format(command, e))


class GraphModelsTests(TestCase):
    """
    Tests for the `graph_models` management command.
    """

    def test_use_model(self):
        include_models = [
            "NoWildcardInclude",
            "Wildcard*InsideInclude",
            "*WildcardPrefixInclude",
            "WildcardSuffixInclude*",
            "*WildcardBothInclude*",
        ]
        exclude_models = [
            "NoWildcardExclude",
            "Wildcard*InsideExclude",
            "*WildcardPrefixExclude",
            "WildcardSuffixExclude*",
            "*WildcardBothExclude*",
            "*Include",
        ]
        # Any model name should be used if neither include or exclude
        # are defined.
        self.assertTrue(
            use_model(
                "SomeModel",
                None,
                None,
            )
        )
        # Any model name should be allowed if `*` is in `include_models`.
        self.assertTrue(
            use_model(
                "SomeModel",
                ["OtherModel", "*", "Wildcard*Model"],
                None,
            )
        )
        # No model name should be allowed if `*` is in `exclude_models`.
        self.assertFalse(
            use_model(
                "SomeModel",
                None,
                ["OtherModel", "*", "Wildcard*Model"],
            )
        )
        # Some tests with the `include_models` defined above.
        self.assertFalse(
            use_model(
                "SomeModel",
                include_models,
                None,
            )
        )
        self.assertTrue(
            use_model(
                "NoWildcardInclude",
                include_models,
                None,
            )
        )
        self.assertTrue(
            use_model(
                "WildcardSomewhereInsideInclude",
                include_models,
                None,
            )
        )
        self.assertTrue(
            use_model(
                "MyWildcardPrefixInclude",
                include_models,
                None,
            )
        )
        self.assertTrue(
            use_model(
                "WildcardSuffixIncludeModel",
                include_models,
                None,
            )
        )
        self.assertTrue(
            use_model(
                "MyWildcardBothIncludeModel",
                include_models,
                None,
            )
        )
        # Some tests with the `exclude_models` defined above.
        self.assertTrue(
            use_model(
                "SomeModel",
                None,
                exclude_models,
            )
        )
        self.assertFalse(
            use_model(
                "NoWildcardExclude",
                None,
                exclude_models,
            )
        )
        self.assertFalse(
            use_model(
                "WildcardSomewhereInsideExclude",
                None,
                exclude_models,
            )
        )
        self.assertFalse(
            use_model(
                "MyWildcardPrefixExclude",
                None,
                exclude_models,
            )
        )
        self.assertFalse(
            use_model(
                "WildcardSuffixExcludeModel",
                None,
                exclude_models,
            )
        )
        self.assertFalse(
            use_model(
                "MyWildcardBothExcludeModel",
                None,
                exclude_models,
            )
        )
        # Test with `exclude_models` and `include_models` combined
        # where the user wants to exclude some models through a wildcard
        # while still being able to include given models
        self.assertTrue(
            use_model("MyWildcardPrefixInclude", include_models, exclude_models)
        )
        self.assertFalse(use_model("MyInclude", include_models, exclude_models))

    def test_no_models_dot_py(self):
        data = generate_graph_data(["testapp_with_no_models_file"])
        self.assertEqual(len(data["graphs"]), 1)

        model_name = data["graphs"][0]["models"][0]["name"]
        self.assertEqual(model_name, "TeslaCar")


class ShowUrlsTests(TestCase):
    """
    Tests for the `show_urls` management command.
    """

    def test_color(self):
        with force_color_support:
            out = StringIO()
            call_command("show_urls", stdout=out)
            self.output = out.getvalue()
            self.assertIn("\x1b", self.output)

    def test_no_color(self):
        with force_color_support:
            out = StringIO()
            call_command("show_urls", "--no-color", stdout=out)
            self.output = out.getvalue()
            self.assertNotIn("\x1b", self.output)


class ListModelInfoTests(TestCase):
    """
    Tests for the `list_model_info` management command.
    """

    def test_plain(self):
        out = StringIO()
        call_command(
            "list_model_info",
            "--model",
            "django_extensions.MultipleFieldsAndMethods",
            stdout=out,
        )
        self.output = out.getvalue()
        self.assertIn("id", self.output)
        self.assertIn("char_field", self.output)
        self.assertIn("integer_field", self.output)
        self.assertIn("foreign_key_field", self.output)
        self.assertIn("has_self_only()", self.output)
        self.assertIn("has_one_extra_argument()", self.output)
        self.assertIn("has_two_extra_arguments()", self.output)
        self.assertIn("has_args_kwargs()", self.output)
        self.assertIn("has_defaults()", self.output)
        self.assertNotIn("__class__()", self.output)
        self.assertNotIn("validate_unique()", self.output)

    def test_all(self):
        out = StringIO()
        call_command(
            "list_model_info",
            "--model",
            "django_extensions.MultipleFieldsAndMethods",
            "--all",
            stdout=out,
        )
        self.output = out.getvalue()
        self.assertIn("id", self.output)
        self.assertIn("__class__()", self.output)
        self.assertIn("validate_unique()", self.output)

    def test_signature(self):
        out = StringIO()
        call_command(
            "list_model_info",
            "--model",
            "django_extensions.MultipleFieldsAndMethods",
            "--signature",
            stdout=out,
        )
        self.output = out.getvalue()
        self.assertIn("has_self_only(self)", self.output)
        self.assertIn("has_one_extra_argument(self, arg_one)", self.output)
        self.assertIn("has_two_extra_arguments(self, arg_one, arg_two)", self.output)
        self.assertIn("has_args_kwargs(self, *args, **kwargs)", self.output)
        self.assertIn(
            "has_defaults(self, one=1, two='Two', true=True, false=False, none=None)",
            self.output,
        )

    def test_db_type(self):
        out = StringIO()
        call_command(
            "list_model_info",
            "--model",
            "django_extensions.MultipleFieldsAndMethods",
            "--db-type",
            stdout=out,
        )
        self.output = out.getvalue()
        self.assertIn("id - integer", self.output)
        self.assertIn("char_field - varchar(10)", self.output)
        self.assertIn("integer_field - integer", self.output)
        self.assertIn("foreign_key_field - integer", self.output)

    def test_field_class(self):
        out = StringIO()
        call_command(
            "list_model_info",
            "--model",
            "django_extensions.MultipleFieldsAndMethods",
            "--field-class",
            stdout=out,
        )
        self.output = out.getvalue()
        self.assertIn("id - AutoField", self.output)
        self.assertIn("char_field - CharField", self.output)
        self.assertIn("integer_field - IntegerField", self.output)
        self.assertIn("foreign_key_field - ForeignKey", self.output)


class MergeModelInstancesTests(TestCase):
    """
    Tests for the `merge_model_instances` management command.
    """

    @mock.patch(
        "django_extensions.management.commands.merge_model_instances.apps.get_models"
    )
    @mock.patch("django_extensions.management.commands.merge_model_instances.input")
    def test_get_model_to_merge(self, test_input, get_models):
        class Model:
            __name__ = ""

        return_value = []
        for v in ["one", "two", "three"]:
            instance = Model()
            instance.__name__ = v
            return_value.append(instance)
        get_models.return_value = return_value
        test_input.return_value = 2
        model_to_deduplicate = get_model_to_deduplicate()
        self.assertEqual(model_to_deduplicate.__name__, "two")

    @mock.patch("django_extensions.management.commands.merge_model_instances.input")
    def test_get_field_names(self, test_input):
        class Field:
            name = ""

            def __init__(self, name):
                self.name = name

        class Model:
            __name__ = ""
            one = Field(name="one")
            two = Field(name="two")
            three = Field(name="three")

        return_value = [
            Model().__getattribute__(field)
            for field in dir(Model())
            if not field.startswith("__")
        ]
        Model._meta = mock.MagicMock()
        Model._meta.get_fields = mock.MagicMock(return_value=return_value)

        # Choose the second return_value
        test_input.side_effect = [2, "C"]
        field_names = get_field_names(Model())
        # Test that the second return_value returned
        self.assertEqual(field_names, [return_value[1].name])

    @mock.patch("django_extensions.management.commands.merge_model_instances.input")
    def test_keep_first_or_last_instance(self, test_input):
        test_input.side_effect = ["xxxx", "first", "last"]
        first_or_last = keep_first_or_last_instance()
        self.assertEqual(first_or_last, "first")
        first_or_last = keep_first_or_last_instance()
        self.assertEqual(first_or_last, "last")

    @mock.patch(
        "django_extensions.management.commands.merge_model_instances.get_model_to_deduplicate"
    )
    @mock.patch(
        "django_extensions.management.commands.merge_model_instances.get_field_names"
    )
    @mock.patch(
        "django_extensions.management.commands.merge_model_instances.keep_first_or_last_instance"
    )
    def test_merge_model_instances(
        self, keep_first_or_last_instance, get_field_names, get_model_to_deduplicate
    ):
        get_model_to_deduplicate.return_value = Person
        get_field_names.return_value = ["name"]
        keep_first_or_last_instance.return_value = "first"

        name = Name.objects.create(name="Name")
        note = Note.objects.create(note="This is a note.")
        personality_1 = Personality.objects.create(description="Child 1's personality.")
        personality_2 = Personality.objects.create(description="Child 2's personality.")
        child_1 = Person.objects.create(
            name=Name.objects.create(name="Child1"),
            age=10,
            personality=personality_1,
        )
        child_1.notes.add(note)
        child_2 = Person.objects.create(
            name=Name.objects.create(name="Child2"),
            age=10,
            personality=personality_2,
        )
        child_2.notes.add(note)

        club1 = Club.objects.create(name="Club one")
        club2 = Club.objects.create(name="Club two")
        person_1 = Person.objects.create(
            name=name,
            age=50,
            personality=Personality.objects.create(description="First personality"),
        )
        person_1.children.add(child_1)
        person_1.notes.add(note)
        Permission.objects.create(text="Permission", person=person_1)

        person_2 = Person.objects.create(
            name=name,
            age=50,
            personality=Personality.objects.create(description="Second personality"),
        )
        person_2.children.add(child_2)
        new_note = Note.objects.create(note="This is a new note")
        person_2.notes.add(new_note)
        Membership.objects.create(club=club1, person=person_2)
        Membership.objects.create(club=club1, person=person_2)
        Permission.objects.create(text="Permission", person=person_2)

        person_3 = Person.objects.create(
            name=name,
            age=50,
            personality=Personality.objects.create(description="Third personality"),
        )
        person_3.children.add(child_2)
        person_3.notes.add(new_note)
        Membership.objects.create(club=club2, person=person_3)
        Membership.objects.create(club=club2, person=person_3)
        Permission.objects.create(text="Permission", person=person_3)

        self.assertEqual(Person.objects.count(), 5)
        self.assertEqual(Membership.objects.count(), 4)
        out = StringIO()
        call_command("merge_model_instances", stdout=out)
        self.ouptput = out.getvalue()
        self.assertEqual(Person.objects.count(), 3)
        person = Person.objects.get(name__name="Name")
        self.assertRaises(
            Person.DoesNotExist,
            lambda: Person.objects.get(personality__description="Second personality"),
        )
        self.assertEqual(person.notes.count(), 2)
        self.assertEqual(person.clubs.distinct().count(), 2)
        self.assertEqual(person.permission_set.count(), 3)
        self.assertRaises(
            Personality.DoesNotExist,
            lambda: Personality.objects.get(description="Second personality"),
        )


class RunJobsTests(TestCase):
    """
    Tests for the `runjobs` management command.
    """

    @mock.patch(
        "django_extensions.management.commands.runjobs.Command.runjobs_by_signals"
    )
    @mock.patch("django_extensions.management.commands.runjobs.Command.runjobs")
    @mock.patch("django_extensions.management.commands.runjobs.Command.usage_msg")
    def test_runjobs_management_command(self, usage_msg, runjobs, runjobs_by_signals):
        when = "daily"
        call_command("runjobs", when)
        usage_msg.assert_not_called()
        runjobs.assert_called_once()
        runjobs_by_signals.assert_called_once()
        self.assertEqual(runjobs.call_args[0][0], when)

    @mock.patch(
        "django_extensions.management.commands.runjobs.Command.runjobs_by_signals"
    )
    @mock.patch("django_extensions.management.commands.runjobs.Command.runjobs")
    @mock.patch("django_extensions.management.commands.runjobs.Command.usage_msg")
    def test_runjobs_management_command_invalid_when(
        self, usage_msg, runjobs, runjobs_by_signals
    ):
        when = "invalid"
        call_command("runjobs", when)
        usage_msg.assert_called_once_with()
        runjobs.assert_not_called()
        runjobs_by_signals.assert_not_called()

    def test_runjobs_integration_test(self):
        jobs = [
            ("hourly", HOURLY_JOB_MOCK),
            ("daily", DAILY_JOB_MOCK),
            ("monthly", MONTHLY_JOB_MOCK),
            ("weekly", WEEKLY_JOB_MOCK),
            ("yearly", YEARLY_JOB_MOCK),
        ]

        # Reset all mocks in case they have been called elsewhere.
        for job in jobs:
            job[1].reset_mock()

        counter = 1
        for job in jobs:
            call_command("runjobs", job[0], verbosity=2)
            for already_called in jobs[:counter]:
                already_called[1].assert_called_once_with()
            for not_yet_called in jobs[counter:]:
                not_yet_called[1].assert_not_called()
            counter += 1

    def test_runjob_integration_test(self):
        jobs = [
            ("test_hourly_job", HOURLY_JOB_MOCK),
            ("test_daily_job", DAILY_JOB_MOCK),
            ("test_monthly_job", MONTHLY_JOB_MOCK),
            ("test_weekly_job", WEEKLY_JOB_MOCK),
            ("test_yearly_job", YEARLY_JOB_MOCK),
        ]

        # Reset all mocks in case they have been called elsewhere.
        for job in jobs:
            job[1].reset_mock()

        counter = 1
        for job in jobs:
            call_command("runjob", job[0], verbosity=2)
            for already_called in jobs[:counter]:
                already_called[1].assert_called_once_with()
            for not_yet_called in jobs[counter:]:
                not_yet_called[1].assert_not_called()
            counter += 1



# -*- coding: utf-8 -*-
import os
import shutil
from tempfile import mkdtemp

from django.test import TestCase

from django_extensions.management.mysql import parse_mysql_cnf


class ParseMysqlCnfTests(TestCase):
    """Tests for parse_mysql_cnf."""

    @classmethod
    def setUpClass(cls):
        cls.tmpdir = mkdtemp()

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmpdir)

    def test_should_return_empty_strings_if_read_default_file_option_is_missing(self):
        dbinfo = {}

        result = parse_mysql_cnf(dbinfo)

        self.assertEqual(result, ("", "", "", "", ""))

    def test_should_parse_my_cnf_and_retun_connection_settings(self):
        my_cnf_path = os.path.join(self.tmpdir, "my.cnf")
        with open(my_cnf_path, "w") as f:
            f.write(
                """[client]
database = test_name
user = test_user
password = test_password
host = localhost
port = 3306
socket = /var/lib/mysqld/mysql.sock
"""
            )

        dbinfo = {
            "ENGINE": "django.db.backends.mysql",
            "OPTIONS": {
                "read_default_file": my_cnf_path,
            },
        }

        result = parse_mysql_cnf(dbinfo)

        self.assertEqual(
            result,
            (
                "test_user",
                "test_password",
                "test_name",
                "/var/lib/mysqld/mysql.sock",
                "3306",
            ),
        )

    def test_should_return_empty_strings_if_NoSectionError_exception_occured(self):
        my_cnf_path = os.path.join(self.tmpdir, "my.cnf")
        with open(my_cnf_path, "w") as f:
            f.write("")

        dbinfo = {
            "ENGINE": "django.db.backends.mysql",
            "OPTIONS": {
                "read_default_file": my_cnf_path,
            },
        }
        result = parse_mysql_cnf(dbinfo)

        self.assertEqual(result, ("", "", "", "", ""))



# -*- coding: utf-8 -*-
from importlib import import_module
from io import StringIO

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.management import CommandError, call_command
from django.test import TestCase

from unittest.mock import patch


class PrintUserForSessionExceptionsTests(TestCase):
    """Test if print_user_for_session command raises exception."""

    def test_should_raise_CommandError_if_session_key_contains_exclamination_mark(self):
        with self.assertRaisesRegex(CommandError, "malformed session key"):
            call_command("print_user_for_session", "l6hxnwblpvrfu8bohelmqjj4soyo2r!?")


class PrintUserForSessionTests(TestCase):
    """Test for print_user_for_session command."""

    def setUp(self):
        self.engine = import_module(settings.SESSION_ENGINE)

    @patch("sys.stdout", new_callable=StringIO)
    def test_should_print_Session_Key_does_not_exist_or_expired(self, m_stdout):
        call_command("print_user_for_session", "l6hxnwblpvrfu8bohelmqjj4soyo2r12")

        self.assertIn("Session Key does not exist. Expired?", m_stdout.getvalue())

    @patch("sys.stdout", new_callable=StringIO)
    def test_should_print_that_there_is_no_user_associated_with_given_session(
        self, m_stdout
    ):
        session = self.engine.SessionStore()
        session.update(
            {
                "_auth_user_backend": "django.contrib.auth.backends.ModelBackend",
                "_auth_user_hash": "b67352fde8582b12f068c10fd9d29f9fa1af0459",
            }
        )
        session.create()

        call_command("print_user_for_session", session.session_key)

        self.assertIn("No user associated with session", m_stdout.getvalue())

    @patch("sys.stdout", new_callable=StringIO)
    def test_should_print_that_there_is_no_backend_associated_with_given_session(
        self, m_stdout
    ):
        session = self.engine.SessionStore()
        session.update(
            {
                "_auth_user_id": 1234,
                "_auth_user_hash": "b67352fde8582b12f068c10fd9d29f9fa1af0459",
            }
        )
        session.create()

        call_command("print_user_for_session", session.session_key)

        self.assertIn(
            "No authentication backend associated with session", m_stdout.getvalue()
        )

    @patch("sys.stdout", new_callable=StringIO)
    def test_should_print_that_there_is_no_user_associated_with_id(self, m_stdout):
        session = self.engine.SessionStore()
        session.update(
            {
                "_auth_user_id": 1234,
                "_auth_user_backend": "django.contrib.auth.backends.ModelBackend",
                "_auth_user_hash": "b67352fde8582b12f068c10fd9d29f9fa1af0459",
            }
        )
        session.create()

        call_command("print_user_for_session", session.session_key)

        self.assertIn("No user associated with that id.", m_stdout.getvalue())

    @patch("sys.stdout", new_callable=StringIO)
    def test_should_print_user_info_for_session(self, m_stdout):
        user = get_user_model().objects.create(
            first_name="John", last_name="Doe", username="foobar", email="foo@bar.com"
        )
        session = self.engine.SessionStore()
        session.update(
            {
                "_auth_user_id": user.pk,
                "_auth_user_backend": "django.contrib.auth.backends.ModelBackend",
                "_auth_user_hash": "b67352fde8582b12f068c10fd9d29f9fa1af0459",
            }
        )
        session.create()
        expected_out = """User id: {}
full name: John Doe
short name: John
username: foobar
email: foo@bar.com
""".format(user.pk)

        call_command("print_user_for_session", session.session_key)

        self.assertIn(expected_out, m_stdout.getvalue())



# -*- coding: utf-8 -*-
import logging
import sys
from io import StringIO

from django.core.management import call_command
from django.test import TestCase

from unittest.mock import patch


class RunJobTests(TestCase):
    def setUp(self):
        sys.stdout = StringIO()
        sys.stderr = StringIO()

        # Remove old loggers, since utils.setup_logger does not clean up after itself
        logger = logging.getLogger("django_extensions.management.commands.runjob")
        for handler in list(logger.handlers):
            logger.removeHandler(handler)

    def test_runs(self):
        # lame test...does it run?
        call_command("runjob", "cache_cleanup", verbosity=2)
        self.assertIn("Executing job: cache_cleanup (app: None)", sys.stdout.getvalue())

    def test_sample_job(self):
        call_command("runjob", "sample_job", verbosity=2)
        self.assertIn("Executing job: sample_job (app: None)", sys.stdout.getvalue())
        self.assertIn("executing empty sample job", sys.stdout.getvalue())

    def test_list_jobs(self):
        call_command("runjob", "-l", verbosity=2)
        self.assertRegex(
            sys.stdout.getvalue(), "tests.testapp +- sample_job +- +- My sample job.\n"
        )

    def test_list_jobs_appconfig(self):
        with self.modify_settings(
            INSTALLED_APPS={
                "append": "tests.testapp.apps.TestAppConfig",
                "remove": "tests.testapp",
            }
        ):
            call_command("runjob", "-l", verbosity=2)
            self.assertRegex(
                sys.stdout.getvalue(),
                "tests.testapp +- sample_job +- +- My sample job.\n",
            )

    def test_runs_appconfig(self):
        with self.modify_settings(
            INSTALLED_APPS={
                "append": "tests.testapp.apps.TestAppConfig",
                "remove": "tests.testapp",
            }
        ):
            call_command("runjob", "sample_job", verbosity=2)
            self.assertIn(
                "Executing job: sample_job (app: None)", sys.stdout.getvalue()
            )
            self.assertIn("executing empty sample job", sys.stdout.getvalue())

    def test_should_print_that_job_not_found(self):
        call_command("runjob", "test_job", verbosity=2)

        self.assertIn("Error: Job test_job not found", sys.stdout.getvalue())

    def test_should_print_that_applabel_not_found(self):
        call_command("runjob", "test_job", "test_app", verbosity=2)

        self.assertIn(
            "Error: Job test_app for applabel test_job not found", sys.stdout.getvalue()
        )

    def test_should_always_print_list_option_usage_if_job_or_applabel_not_found(self):
        call_command("runjob", "test_job", verbosity=2)

        self.assertIn(
            "Use -l option to view all the available jobs", sys.stdout.getvalue()
        )

    @patch("django_extensions.management.commands.runjob.get_job")
    def test_should_print_traceback(self, m_get_job):
        m_get_job.return_value.return_value.execute.side_effect = Exception

        call_command("runjob", "test_job", "test_app")

        self.assertIn(
            "ERROR OCCURED IN JOB: test_app (APP: test_job)", sys.stdout.getvalue()
        )
        self.assertIn("Traceback (most recent call last):", sys.stdout.getvalue())



# -*- coding: utf-8 -*-
from typing import Optional, Set  # noqa

from django.conf import settings
from django.test.utils import override_settings

from tests.management.commands.shell_plus_tests.test_utils import (
    AutomaticShellPlusImportsTestCase,
)
from tests.test_module_in_project_dir import FourthDerivedClass
from tests.testapp.derived_classes_for_testing import SecondDerivedClass
from tests.testapp.derived_classes_for_testing.test_module import (
    ClassWhichShouldNotBeImported,
    ThirdDerivedClass,
)
from tests.testapp.classes_to_include import (
    BaseIncludedClass,
    FirstDerivedClass,
    IncludedMixin,
)


class ImportSubclassesTestCase(AutomaticShellPlusImportsTestCase):
    def test_imports_no_subclasses(self):
        self.assert_imports()

    @override_settings(
        SHELL_PLUS_SUBCLASSES_IMPORT=[],
    )
    def test_imports_empty_list(self):
        self.assert_imports()

    @override_settings(
        SHELL_PLUS_SUBCLASSES_IMPORT=[BaseIncludedClass],
    )
    def test_imports_one_base_class(self):
        self.assert_imports(
            first={"FirstDerivedClass"},
            second={"SecondDerivedClass"},
            fourth={"FourthDerivedClass"},
        )

    @override_settings(
        SHELL_PLUS_SUBCLASSES_IMPORT=[
            "tests.testapp.classes_to_include.BaseIncludedClass"
        ],
    )
    def test_imports_one_base_class_as_string(self):
        self.assert_imports(
            first={"FirstDerivedClass"},
            second={"SecondDerivedClass"},
            fourth={"FourthDerivedClass"},
        )

    @override_settings(
        SHELL_PLUS_SUBCLASSES_IMPORT=[IncludedMixin],
    )
    def test_imports_one_base_mixin(self):
        self.assert_imports(
            first={"FirstDerivedClass"},
            third={"ThirdDerivedClass"},
        )

    @override_settings(
        SHELL_PLUS_SUBCLASSES_IMPORT=[BaseIncludedClass, IncludedMixin],
    )
    def test_imports_two_base_classes(self):
        self.assert_imports(
            first={"FirstDerivedClass"},
            second={"SecondDerivedClass"},
            third={"ThirdDerivedClass"},
            fourth={"FourthDerivedClass"},
        )

    @override_settings(
        SHELL_PLUS_SUBCLASSES_IMPORT=[BaseIncludedClass, IncludedMixin],
        SHELL_PLUS_SUBCLASSES_IMPORT_MODULES_BLACKLIST=settings.SHELL_PLUS_SUBCLASSES_IMPORT_MODULES_BLACKLIST
        + [
            "tests.testapp",
        ],
    )
    def test_imports_two_base_classes_exclude_testapp(self):
        self.assert_imports(
            fourth={"FourthDerivedClass"},
        )

    @override_settings(
        SHELL_PLUS_SUBCLASSES_IMPORT=[BaseIncludedClass, IncludedMixin],
        SHELL_PLUS_SUBCLASSES_IMPORT_MODULES_BLACKLIST=settings.SHELL_PLUS_SUBCLASSES_IMPORT_MODULES_BLACKLIST
        + [
            "tests.testapp.derived_classes_for_testing",
        ],
    )
    def test_imports_two_base_classes_exclude_derived_class_for_testing(self):
        self.assert_imports(
            first={"FirstDerivedClass"},
            fourth={"FourthDerivedClass"},
        )

    @override_settings(
        SHELL_PLUS_SUBCLASSES_IMPORT=[BaseIncludedClass, IncludedMixin],
        SHELL_PLUS_SUBCLASSES_IMPORT_MODULES_BLACKLIST=settings.SHELL_PLUS_SUBCLASSES_IMPORT_MODULES_BLACKLIST
        + [
            "tests.testapp.derived_classes_for_testing.test_module",
        ],
    )
    def test_imports_two_base_classes_exclude_test_module(self):
        self.assert_imports(
            first={"FirstDerivedClass"},
            second={"SecondDerivedClass"},
            fourth={"FourthDerivedClass"},
        )

    @override_settings(
        SHELL_PLUS_SUBCLASSES_IMPORT=[BaseIncludedClass, IncludedMixin],
        SHELL_PLUS_SUBCLASSES_IMPORT_MODULES_BLACKLIST=settings.SHELL_PLUS_SUBCLASSES_IMPORT_MODULES_BLACKLIST
        + [
            "tests.test_module_in_project_dir",
        ],
    )
    def test_imports_two_base_classes_exclude_classes_in_project_dir(self):
        self.assert_imports(
            first={"FirstDerivedClass"},
            second={"SecondDerivedClass"},
            third={"ThirdDerivedClass"},
        )

    @override_settings(
        SHELL_PLUS_SUBCLASSES_IMPORT=[BaseIncludedClass, IncludedMixin],
        SHELL_PLUS_SUBCLASSES_IMPORT_MODULES_BLACKLIST=settings.SHELL_PLUS_SUBCLASSES_IMPORT_MODULES_BLACKLIST
        + [
            "tests.testapp.classes_to_include",
        ],
    )
    def test_imports_two_base_classes_exclude_classes_in_classes_to_include(self):
        self.assert_imports(
            second={"SecondDerivedClass"},
            third={"ThirdDerivedClass"},
            fourth={"FourthDerivedClass"},
        )

    def assert_imports(self, first=None, second=None, third=None, fourth=None):
        """
        Auxiliary assertion which checks are classes imported under names.
        :param first: set of expected names under which FirstDerivedClass should be available
        :param second: set of expected names under which SecondDerivedClass should be available
        :param third: set of expected names under which ThirdDerivedClass should be available
        :param fourth: set of expected names under which FourthDerivedClass should be available
        """
        # type: (Optional[Set[str]], Optional[Set[str]], Optional[Set[str]], Optional[Set[str]]) -> ()
        self.run_shell_plus()
        self.assert_imported_under_names(FirstDerivedClass, first or set())
        self.assert_imported_under_names(SecondDerivedClass, second or set())
        self.assert_imported_under_names(ThirdDerivedClass, third or set())
        self.assert_imported_under_names(FourthDerivedClass, fourth or set())
        self.assert_imported_under_names(ClassWhichShouldNotBeImported, set())



# -*- coding: utf-8 -*-
import os
import sys
import importlib

from io import StringIO
from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase, override_settings

from django_extensions.management.commands.runscript import (
    Command,
    BadCustomDirectoryException,
    DirPolicyChoices,
)


class RunScriptTests(TestCase):
    def setUp(self):
        sys.stdout = StringIO()
        sys.stderr = StringIO()

    def get_command(self):
        cmd = Command()
        cmd.running_tests = True
        return cmd

    def test_runs(self):
        # lame test...does it run?
        call_command("runscript", "sample_script", verbosity=2)
        self.assertIn(
            "Found script 'tests.testapp.scripts.sample_script'", sys.stdout.getvalue()
        )
        self.assertIn(
            "Running script 'tests.testapp.scripts.sample_script'",
            sys.stdout.getvalue(),
        )

    def test_runs_appconfig(self):
        with self.modify_settings(
            INSTALLED_APPS={
                "append": "tests.testapp.apps.TestAppConfig",
                "remove": "tests.testapp",
            }
        ):
            call_command("runscript", "sample_script", verbosity=2)
            self.assertIn(
                "Found script 'tests.testapp.scripts.sample_script'",
                sys.stdout.getvalue(),
            )
            self.assertIn(
                "Running script 'tests.testapp.scripts.sample_script'",
                sys.stdout.getvalue(),
            )


class NonExistentScriptsTests(RunScriptTests):
    def test_prints_error_on_nonexistent_script(self):
        cmd = self.get_command()
        with self.assertRaises(CommandError):
            call_command(cmd, "non_existent_script", verbosity=2)
        self.assertIn(
            "No (valid) module for script 'non_existent_script' found",
            sys.stdout.getvalue(),
        )
        self.assertEqual(cmd.last_exit_code, 1)

    def test_prints_nothing_for_nonexistent_script_when_silent(self):
        cmd = self.get_command()
        call_command(cmd, "non_existent_script", silent=True)
        self.assertEqual("", sys.stdout.getvalue())
        self.assertEqual(cmd.last_exit_code, 1)

    def test_doesnt_print_exception_for_nonexistent_script_when_no_traceback(self):
        cmd = self.get_command()
        with self.assertRaises(CommandError):
            call_command(cmd, "non_existent_script", no_traceback=True)
        self.assertEqual("", sys.stderr.getvalue())
        self.assertEqual(cmd.last_exit_code, 1)


class InvalidImportScriptsTests(RunScriptTests):
    def test_prints_additional_info_on_nonexistent_script_by_default(self):
        cmd = self.get_command()
        with self.assertRaises(CommandError):
            call_command(cmd, "non_existent_script")
        self.assertIn(
            "No (valid) module for script 'non_existent_script' found",
            sys.stdout.getvalue(),
        )
        self.assertIn(
            "Try running with a higher verbosity level like: -v2 or -v3",
            sys.stdout.getvalue(),
        )
        self.assertEqual(cmd.last_exit_code, 1)

    def test_prints_import_error_on_script_with_invalid_imports_by_default(self):
        cmd = self.get_command()
        with self.assertRaises(CommandError):
            call_command(cmd, "invalid_import_script")
        self.assertIn(
            "Cannot import module 'tests.testapp.scripts.invalid_import_script'",
            sys.stdout.getvalue(),
        )
        self.assertRegex(
            sys.stdout.getvalue(), "No module named (')?(invalidpackage)\1?"
        )
        self.assertEqual(cmd.last_exit_code, 1)

    def test_prints_import_error_on_script_with_invalid_imports_reliably(self):
        cmd = self.get_command()
        if hasattr(importlib, "util") and hasattr(importlib.util, "find_spec"):
            with self.settings(BASE_DIR=os.path.dirname(os.path.abspath(__file__))):
                with self.assertRaises(CommandError):
                    call_command(cmd, "invalid_import_script")
            self.assertIn(
                "Cannot import module 'tests.testapp.scripts.invalid_import_script'",
                sys.stdout.getvalue(),
            )
            self.assertRegex(
                sys.stdout.getvalue(), "No module named (')?(invalidpackage)\1?"
            )
            self.assertEqual(cmd.last_exit_code, 1)


class InvalidScriptsTests(RunScriptTests):
    def test_raises_error_message_on_invalid_script_by_default(self):
        cmd = self.get_command()
        with self.assertRaises(Exception):
            call_command(cmd, "error_script")
        self.assertIn("Exception while running run() in", sys.stdout.getvalue())

    def test_prints_nothing_for_invalid_script_when_silent(self):
        cmd = self.get_command()
        call_command(cmd, "error_script", silent=True)
        self.assertEqual(cmd.last_exit_code, 1)
        self.assertEqual("", sys.stdout.getvalue())

    def test_doesnt_print_exception_for_nonexistent_script_when_no_traceback(self):
        cmd = self.get_command()
        with self.assertRaises(CommandError):
            call_command(cmd, "error_script", no_traceback=True)
        self.assertEqual("", sys.stderr.getvalue())
        self.assertIn("Exception while running run() in", sys.stdout.getvalue())
        self.assertEqual(cmd.last_exit_code, 1)


class RunFunctionTests(RunScriptTests):
    def test_prints_error_message_for_script_without_run(self):
        cmd = self.get_command()
        with self.assertRaises(CommandError):
            call_command(cmd, "script_no_run_function")
        self.assertIn(
            "No (valid) module for script 'script_no_run_function' found",
            sys.stdout.getvalue(),
        )
        self.assertIn(
            "Try running with a higher verbosity level like: -v2 or -v3",
            sys.stdout.getvalue(),
        )
        self.assertEqual(cmd.last_exit_code, 1)

    def test_prints_additional_info_for_script__run_extra_verbosity(self):
        cmd = self.get_command()
        with self.assertRaises(CommandError):
            call_command(cmd, "script_no_run_function", verbosity=2)
        self.assertIn(
            "No (valid) module for script 'script_no_run_function' found",
            sys.stdout.getvalue(),
        )
        self.assertIn("Found script", sys.stdout.getvalue())
        self.assertEqual(cmd.last_exit_code, 1)

    def test_prints_nothing_for_script_without_run(self):
        cmd = self.get_command()
        call_command(cmd, "script_no_run_function", silent=True)
        self.assertEqual("", sys.stdout.getvalue())


project_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))


class ChangingDirectoryTests(RunScriptTests):
    def setUp(self):
        super().setUp()
        self.curwd = os.getcwd()
        os.chdir(project_path)

    def tearDown(self):
        super().setUp()
        os.chdir(self.curwd)

    def _execute_script_with_chdir(
        self, dir_policy, start_path, expected_path, chdir=None
    ):
        os.chdir(os.path.join(project_path, *start_path))
        expected_path = os.path.join(project_path, *expected_path)
        call_command(
            "runscript", "directory_checker_script", dir_policy=dir_policy, chdir=chdir
        )
        output = sys.stdout.getvalue().split("Script called from: ")[1]
        self.assertEqual(output, expected_path + "\n")

    def test_none_policy_command_run(self):
        self._execute_script_with_chdir(DirPolicyChoices.NONE, [], [])

    def test_none_policy_command_run_with_chdir(self):
        self._execute_script_with_chdir(DirPolicyChoices.NONE, ["tests"], ["tests"])

    def test_none_policy_freezing_start_directory(self):
        self._execute_script_with_chdir(DirPolicyChoices.NONE, ["tests"], ["tests"])
        self._execute_script_with_chdir(DirPolicyChoices.NONE, ["tests"], ["tests"])

    def test_root_policy_command_run(self):
        self._execute_script_with_chdir(DirPolicyChoices.ROOT, ["tests"], [])

    def test_each_policy_command_run(self):
        os.chdir(os.path.join(project_path, "tests"))
        call_command(
            "runscript",
            "directory_checker_script",
            "other_directory_checker_script",
            dir_policy=DirPolicyChoices.EACH,
        )
        output = sys.stdout.getvalue()
        first_output = output.split("Script called from: ")[1].split(
            "Cannot import module "
        )[0]
        self.assertEqual(
            first_output,
            os.path.join(project_path, "tests", "testapp", "scripts") + "\n",
        )
        second_output = output.split("Script called from: ")[2].split(
            "Cannot import module "
        )[0]
        self.assertEqual(
            second_output,
            os.path.join(
                project_path, "tests", "testapp_with_no_models_file", "scripts"
            )
            + "\n",
        )

    def test_chdir_specified(self):
        execution_path = os.path.join(project_path, "django_extensions", "management")
        self._execute_script_with_chdir(
            DirPolicyChoices.ROOT,
            ["tests"],
            ["django_extensions", "management"],
            chdir=execution_path,
        )

    @override_settings(RUNSCRIPT_CHDIR=os.path.join(project_path, "tests"))
    def test_policy_from_cli_and_chdir_from_settings(self):
        self._execute_script_with_chdir(DirPolicyChoices.ROOT, ["tests"], [])

    @override_settings(
        RUNSCRIPT_CHDIR=os.path.join(project_path, "django_extensions", "management"),
        RUNSCRIPT_CHDIR_POLICY=DirPolicyChoices.ROOT,
    )
    def test_chdir_from_settings_and_policy_from_settings(self):
        self._execute_script_with_chdir(
            None, ["tests"], ["django_extensions", "management"]
        )

    @override_settings(RUNSCRIPT_CHDIR_POLICY=DirPolicyChoices.EACH)
    def test_policy_from_settings(self):
        self._execute_script_with_chdir(
            None, ["tests"], ["tests", "testapp", "scripts"]
        )

    @override_settings(RUNSCRIPT_CHDIR=os.path.join(project_path, "tests"))
    def test_chdir_django_settings(self):
        self._execute_script_with_chdir(None, [], ["tests"])

    @override_settings(RUNSCRIPT_CHDIR="bad path")
    def test_custom_policy_django_settings_bad_path(self):
        with self.assertRaisesRegex(
            BadCustomDirectoryException,
            "bad path is not a directory! If --dir-policy is custom than you must set "
            "correct directory in --dir option or in settings.RUNSCRIPT_CHDIR",
        ):
            self._execute_script_with_chdir(None, [], ["tests"])

    def test_skip_printing_modules_which_does_not_exist(self):
        call_command("runscript", "directory_checker_script")
        self.assertNotIn("No module named", sys.stdout.getvalue())
        self.assertNotIn("No module named", sys.stderr.getvalue())



# -*- coding: utf-8 -*-
from django.test import TestCase
from django.test.utils import override_settings

from django_extensions.compat import get_template_setting


class CompatTests(TestCase):
    @override_settings(TEMPLATES=None)
    def test_should_return_None_by_default_if_TEMPLATES_setting_is_none(self):
        self.assertIsNone(get_template_setting("template_key"))

    @override_settings(TEMPLATES=None)
    def test_should_return_default_if_TEMPLATES_setting_is_none(self):
        self.assertEqual(get_template_setting("template_key", "test"), "test")

    @override_settings(
        TEMPLATES=[
            {
                "BACKEND": "django.template.backends.django.DjangoTemplates",
                "DIRS": ["templates"],
                "APP_DIRS": True,
            }
        ]
    )
    def test_should_return_value_for_key(self):
        self.assertEqual(
            get_template_setting("BACKEND"),
            "django.template.backends.django.DjangoTemplates",
        )



# -*- coding: utf-8 -*-
"""
SyncData
========

Django command similar to 'loaddata' but also deletes.
After 'syncdata' has run, the database will have the same data as the fixture - anything
missing will of been added, anything different will of been updated,
and anything extra will of been deleted.
"""

import os

from django.apps import apps
from django.conf import settings
from django.core import serializers
from django.core.management.base import BaseCommand, CommandError
from django.core.management.color import no_style
from django.db import DEFAULT_DB_ALIAS, connections, transaction
from django.template.defaultfilters import pluralize

from django_extensions.management.utils import signalcommand


def humanize(dirname):
    return "'%s'" % dirname if dirname else "absolute path"


class SyncDataError(Exception):
    pass


class Command(BaseCommand):
    """syncdata command"""

    help = "Makes the current database have the same data as the fixture(s), no more, no less."  # noqa: E501
    args = "fixture [fixture ...]"

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--skip-remove",
            action="store_false",
            dest="remove",
            default=True,
            help="Avoid remove any object from db",
        )
        parser.add_argument(
            "--remove-before",
            action="store_true",
            dest="remove_before",
            default=False,
            help="Remove existing objects before inserting and updating new ones",
        )
        parser.add_argument(
            "--database",
            default=DEFAULT_DB_ALIAS,
            help=(
                "Nominates a specific database to load fixtures into. "
                'Defaults to the "default" database.'
            ),
        )
        parser.add_argument(
            "fixture_labels",
            nargs="?",
            type=str,
            help="Specify the fixture label (comma separated)",
        )

    def remove_objects_not_in(self, objects_to_keep, verbosity):
        """
        Delete all the objects in the database that are not in objects_to_keep.
        - objects_to_keep: A map where the keys are classes, and the values are a
         set of the objects of that class we should keep.
        """
        for class_ in objects_to_keep.keys():
            current = class_.objects.all()
            current_ids = set(x.pk for x in current)
            keep_ids = set(x.pk for x in objects_to_keep[class_])

            remove_these_ones = current_ids.difference(keep_ids)
            if remove_these_ones:
                for obj in current:
                    if obj.pk in remove_these_ones:
                        obj.delete()
                        if verbosity >= 2:
                            print("Deleted object: %s" % str(obj))

            if verbosity > 0 and remove_these_ones:
                num_deleted = len(remove_these_ones)
                if num_deleted > 1:
                    type_deleted = str(class_._meta.verbose_name_plural)
                else:
                    type_deleted = str(class_._meta.verbose_name)

                print("Deleted %s %s" % (str(num_deleted), type_deleted))

    @signalcommand
    def handle(self, *args, **options):
        self.style = no_style()
        self.using = options["database"]
        fixture_labels = (
            options["fixture_labels"].split(",") if options["fixture_labels"] else ()
        )
        try:
            with transaction.atomic():
                self.syncdata(fixture_labels, options)
        except SyncDataError as exc:
            raise CommandError(exc)
        finally:
            # Close the DB connection -- unless we're still in a transaction. This
            # is required as a workaround for an edge case in MySQL: if the same
            # connection is used to create tables, load data, and query, the query
            # can return incorrect results. See Django #7572, MySQL #37735.
            if transaction.get_autocommit(self.using):
                connections[self.using].close()

    def syncdata(self, fixture_labels, options):
        verbosity = options["verbosity"]
        show_traceback = options["traceback"]

        # Keep a count of the installed objects and fixtures
        fixture_count = 0
        object_count = 0
        objects_per_fixture = []
        models = set()

        # Get a cursor (even though we don't need one yet). This has
        # the side effect of initializing the test database (if
        # it isn't already initialized).
        cursor = connections[self.using].cursor()

        app_modules = [app.module for app in apps.get_app_configs()]
        app_fixtures = [
            os.path.join(os.path.dirname(app.__file__), "fixtures")
            for app in app_modules
        ]
        for fixture_label in fixture_labels:
            parts = fixture_label.split(".")
            if len(parts) == 1:
                fixture_name = fixture_label
                formats = serializers.get_public_serializer_formats()
            else:
                fixture_name, format_ = ".".join(parts[:-1]), parts[-1]
                if format_ in serializers.get_public_serializer_formats():
                    formats = [format_]
                else:
                    formats = []

            if formats:
                if verbosity > 1:
                    print("Loading '%s' fixtures..." % fixture_name)
            else:
                raise SyncDataError(
                    (
                        "Problem installing fixture '%s': %s is not a known "
                        "serialization format."
                    )
                    % (fixture_name, format_)
                )

            if os.path.isabs(fixture_name):
                fixture_dirs = [fixture_name]
            else:
                fixture_dirs = app_fixtures + list(settings.FIXTURE_DIRS) + [""]

            for fixture_dir in fixture_dirs:
                if verbosity > 1:
                    print("Checking %s for fixtures..." % humanize(fixture_dir))

                label_found = False
                for format_ in formats:
                    if verbosity > 1:
                        print(
                            "Trying %s for %s fixture '%s'..."
                            % (humanize(fixture_dir), format_, fixture_name)
                        )
                    try:
                        full_path = os.path.join(
                            fixture_dir, ".".join([fixture_name, format_])
                        )
                        fixture = open(full_path, "r")
                        if label_found:
                            fixture.close()
                            raise SyncDataError(
                                "Multiple fixtures named '%s' in %s. Aborting."
                                % (fixture_name, humanize(fixture_dir))
                            )
                        else:
                            fixture_count += 1
                            objects_per_fixture.append(0)
                            if verbosity > 0:
                                print(
                                    "Installing %s fixture '%s' from %s."
                                    % (format_, fixture_name, humanize(fixture_dir))
                                )
                            try:
                                objects_to_keep = {}
                                objects = list(
                                    serializers.deserialize(format_, fixture)
                                )
                                for obj in objects:
                                    class_ = obj.object.__class__
                                    if class_ not in objects_to_keep:
                                        objects_to_keep[class_] = set()
                                    objects_to_keep[class_].add(obj.object)

                                if options["remove"] and options["remove_before"]:
                                    self.remove_objects_not_in(
                                        objects_to_keep, verbosity
                                    )

                                for obj in objects:
                                    object_count += 1
                                    objects_per_fixture[-1] += 1
                                    models.add(obj.object.__class__)
                                    obj.save()

                                if options["remove"] and not options["remove_before"]:
                                    self.remove_objects_not_in(
                                        objects_to_keep, verbosity
                                    )

                                label_found = True
                            except (SystemExit, KeyboardInterrupt):
                                raise
                            except Exception:
                                import traceback

                                fixture.close()
                                if show_traceback:
                                    traceback.print_exc()
                                raise SyncDataError(
                                    "Problem installing fixture '%s': %s\n"
                                    % (full_path, traceback.format_exc())
                                )

                            fixture.close()
                    except SyncDataError as e:
                        raise e
                    except Exception:
                        if verbosity > 1:
                            print(
                                "No %s fixture '%s' in %s."
                                % (format_, fixture_name, humanize(fixture_dir))
                            )

        # If any of the fixtures we loaded contain 0 objects, assume that an
        # error was encountered during fixture loading.
        if 0 in objects_per_fixture:
            raise SyncDataError(
                "No fixture data found for '%s'. (File format may be invalid.)"
                % fixture_name
            )

        # If we found even one object in a fixture, we need to reset the
        # database sequences.
        if object_count > 0:
            sequence_sql = connections[self.using].ops.sequence_reset_sql(
                self.style, models
            )
            if sequence_sql:
                if verbosity > 1:
                    print("Resetting sequences")
                for line in sequence_sql:
                    cursor.execute(line)

        if object_count == 0:
            if verbosity > 1:
                print("No fixtures found.")
        else:
            if verbosity > 0:
                print(
                    "Installed %d object%s from %d fixture%s"
                    % (
                        object_count,
                        pluralize(object_count),
                        fixture_count,
                        pluralize(fixture_count),
                    )
                )



# -*- coding: utf-8 -*-
import logging

from django.core.management.base import BaseCommand

from django_extensions.management.jobs import get_job, print_jobs
from django_extensions.management.utils import setup_logger, signalcommand

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Run a single maintenance job."
    missing_args_message = "test"

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument("app_name", nargs="?")
        parser.add_argument("job_name", nargs="?")
        parser.add_argument(
            "--list",
            "-l",
            action="store_true",
            dest="list_jobs",
            default=False,
            help="List all jobs with their description",
        )

    def runjob(self, app_name, job_name, options):
        verbosity = options["verbosity"]
        if verbosity > 1:
            logger.info("Executing job: %s (app: %s)", job_name, app_name)
        try:
            job = get_job(app_name, job_name)
        except KeyError:
            if app_name:
                logger.error(
                    "Error: Job %s for applabel %s not found", job_name, app_name
                )
            else:
                logger.error("Error: Job %s not found", job_name)
            logger.info("Use -l option to view all the available jobs")
            return
        try:
            job().execute()
        except Exception:
            logger.exception("ERROR OCCURED IN JOB: %s (APP: %s)", job_name, app_name)

    @signalcommand
    def handle(self, *args, **options):
        app_name = options["app_name"]
        job_name = options["job_name"]

        # hack since we are using job_name nargs='?' for -l to work
        if app_name and not job_name:
            job_name = app_name
            app_name = None

        setup_logger(logger, self.stdout)

        if options["list_jobs"]:
            print_jobs(only_scheduled=False, show_when=True, show_appname=True)
        else:
            self.runjob(app_name, job_name, options)



# -*- coding: utf-8 -*-

import pytest
from django.core.management import CommandError, call_command


def test_without_args(capsys):
    call_command("print_settings")

    out, err = capsys.readouterr()
    assert "DEBUG" in out
    assert "INSTALLED_APPS" in out


def test_with_setting_args(capsys):
    call_command("print_settings", "DEBUG")

    out, err = capsys.readouterr()
    assert "DEBUG" in out
    assert "INSTALLED_APPS" not in out


def test_with_setting_wildcard(capsys):
    call_command("print_settings", "*_DIRS")

    out, err = capsys.readouterr()
    assert "FIXTURE_DIRS" in out
    assert "STATICFILES_DIRS" in out
    assert "INSTALLED_APPS" not in out


def test_with_setting_fail(capsys):
    with pytest.raises(CommandError, match="INSTALLED_APPZ not found in settings."):
        call_command("print_settings", "-f", "INSTALLED_APPZ")


def test_with_multiple_setting_args(capsys):
    call_command(
        "print_settings",
        "SECRET_KEY",
        "DATABASES",
        "INSTALLED_APPS",
    )

    out, err = capsys.readouterr()
    assert "DEBUG" not in out
    assert "SECRET_KEY" in out
    assert "DATABASES" in out
    assert "INSTALLED_APPS" in out


def test_format(capsys):
    call_command(
        "print_settings",
        "DEBUG",
        "--format=text",
    )

    out, err = capsys.readouterr()
    expected = "DEBUG = False\n"
    assert expected == out


def test_format_json_without_indent(capsys):
    call_command(
        "print_settings",
        "DEBUG",
        "--format=json",
        "--indent=0",
    )

    expected = '{\n"DEBUG": false\n}\n'
    out, err = capsys.readouterr()
    assert expected == out
